# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings.units.base import _UNIT_TYPES

import mididings.patch as _patch
import mididings.arguments as _arguments


class Scene(object):
    """
    Scene(name, patch, init_patch=None, exit_patch=None)

    Construct a Scene object to be used with the :func:`run()` function.

    :param name: a string describing the scene.
    :param patch: the patch defining the MIDI processing to take place for
        incoming events.
    :param init_patch: an optional patch that will be triggered when
        switching to this scene.
    :param exit_patch: an optional patch that will be triggered when
        switching away from this scene.
    """
    @_arguments.accept(None, _arguments.nullable(str), _UNIT_TYPES,
                       _arguments.nullable(_UNIT_TYPES),
                       _arguments.nullable(_UNIT_TYPES))
    def __init__(self, name, patch, init_patch=None, exit_patch=None):
        self.name = name if name else ''
        self.patch = patch
        self.init_patch = [init_patch] if init_patch else []
        self.exit_patch = [exit_patch] if exit_patch else []


class SceneGroup(object):
    """
    SceneGroup(name, subscenes)

    Construct a SceneGroup object. This can be used to group multiple scenes
    under a common name and program number.

    :param name: a string describing the subscene.
    :param subscenes: a list of :class:`Scene` objects or patches.
    """
    @_arguments.accept(None, str, [(Scene,) + _UNIT_TYPES])
    def __init__(self, name, subscenes):
        self.name = name
        self.subscenes = subscenes



def _parse_scene(scene):
    if isinstance(scene, Scene):
        pass
    elif isinstance(scene, tuple):
        scene = Scene(None, *scene)
    else:
        scene = Scene(None, scene, None)

    # add any initializations defined in the processing patch
    # (via Init() etc.) to the init patch
    scene.init_patch += _patch.get_init_patches(scene.patch)
    scene.exit_patch += _patch.get_exit_patches(scene.patch)

    return scene


# the types that are (convertible to) valid scenes
_SCENE_TYPES = (Scene, SceneGroup) + _UNIT_TYPES



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2011  Sébastien Devaux
# Copyright (C) 2012-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings import *
from mididings import event
from mididings import util


class _CtrlToSysEx(object):
    def __init__(self, sysex, index, checksum_start):
        self.sysex = util.sysex_data(sysex)
        self.index = index
        self.checksum_start = checksum_start

    def __call__(self, ev):
        sysex = self.sysex
        sysex[self.index] = ev.value

        if self.checksum_start is not None:
            checksum = sum(self.sysex[self.checksum_start:-2])
            sysex[-2] = (128 - checksum) % 128

        return event.SysExEvent(ev.port, sysex)


def CtrlToSysEx(ctrl, sysex, index, checksum_start=None):
    """
    Convert control change to system exclusive event using a sysex pattern.

    :param ctrl: controller number to be replaced.
    :param sysex: sysex pattern.
    :param index: index of sysex pattern byte to be replaced by the
        controller value.
    :param checksum_start: index of first sysex byte to be incorporated into
        Roland checksum computation. The checksum will replace the
        next-to-last byte of the sysex pattern.
        If this parameter is ``None``, no checksum is computed.
    """
    return CtrlFilter(ctrl) % Process(
                _CtrlToSysEx(sysex, index, checksum_start))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings.units.base import *
from mididings.units.engine import *
from mididings.units.filters import *
from mididings.units.splits import *
from mididings.units.modifiers import *
from mididings.units.generators import *
from mididings.units.call import *
from mididings.units.printing import *
from mididings.units.init import *


import mididings.misc as _misc
__all__ = _misc.prune_globals(globals())



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings.units.base import Chain, Fork, _UNIT_TYPES
from mididings.units.filters import (
        PortFilter, ChannelFilter, KeyFilter, VelocityFilter,
        CtrlFilter, CtrlValueFilter, ProgramFilter, SysExFilter)

import mididings.overload as _overload
import mididings.arguments as _arguments
import mididings.util as _util


def _make_split(t, d, unpack=False):
    if unpack:
        # if dictionary key is a tuple, unpack and pass as individual
        # parameters to ctor
        t = lambda p, t=t: t(*(p if isinstance(p, tuple) else (p,)))

    # build dict with all items from d, except d[None]
    dd = dict((k, v) for k, v in d.items() if k is not None)

    # build fork from all normal items
    r = Fork((t(k) >> w) for k, w in dd.items())

    # add else-rule, if any
    if None in d:
        f = Chain(~t(k) for k in dd.keys())
        r.append(f >> d[None])

    return r


def _make_threshold(f, patch_lower, patch_upper):
    return Fork([
        f >> patch_lower,
        ~f >> patch_upper,
    ])



@_arguments.accept({
    _arguments.nullable(_arguments.flatten(_util.port_number, tuple)):
        _UNIT_TYPES
})
def PortSplit(mapping):
    """
    PortSplit(mapping)

    Split events by input port, with *mapping* being a dictionary of the form
    ``{ports: patch, ...}``.
    """
    return _make_split(PortFilter, mapping)


@_arguments.accept({
    _arguments.nullable(_arguments.flatten(_util.channel_number, tuple)):
        _UNIT_TYPES
})
def ChannelSplit(mapping):
    """
    ChannelSplit(mapping)

    Split events by input channel, with *mapping* being a dictionary of
    the form ``{channels: patch, ...}``.
    """
    return _make_split(ChannelFilter, mapping)


@_overload.mark(
    """
    KeySplit(threshold, patch_lower, patch_upper)
    KeySplit(mapping)

    Split events by key. Non-note events are sent to all patches.

    The first version splits at a single threshold. The second version allows
    an arbitrary number of (possibly overlapping) note ranges, with *mapping*
    being a dictionary of the form ``{note_range: patch, ...}``.
    """
)
@_arguments.accept(_util.note_limit, _UNIT_TYPES, _UNIT_TYPES)
def KeySplit(threshold, patch_lower, patch_upper):
    return _make_threshold(KeyFilter(0, threshold),
                           patch_lower, patch_upper)

@_overload.mark
@_arguments.accept({_arguments.nullable(_util.note_range): _UNIT_TYPES})
def KeySplit(mapping):
    return _make_split(KeyFilter, mapping)


@_overload.mark(
    """
    VelocitySplit(threshold, patch_lower, patch_upper)
    VelocitySplit(mapping)

    Split events by note-on velocity. Non-note events are sent to all patches.

    The first version splits at a single threshold. The second version allows
    an arbitrary number of (possibly overlapping) value ranges, with
    *mapping* being a dictionary of the form ``{(lower, upper): patch, ...}``.
    """
)
@_arguments.accept(_util.velocity_limit, _UNIT_TYPES, _UNIT_TYPES)
def VelocitySplit(threshold, patch_lower, patch_upper):
    return _make_threshold(VelocityFilter(0, threshold),
                           patch_lower, patch_upper)

@_overload.mark
@_arguments.accept({_arguments.nullable(_util.velocity_range): _UNIT_TYPES})
def VelocitySplit(mapping):
    return _make_split(VelocityFilter, mapping, unpack=True)


@_arguments.accept({
    _arguments.nullable(_arguments.flatten(_util.ctrl_number, tuple)):
        _UNIT_TYPES
})
def CtrlSplit(mapping):
    """
    CtrlSplit(mapping)

    Split events by controller number, with *mapping* being a dictionary of
    the form ``{ctrls: patch, ...}``.
    Non-control-change events are discarded.
    """
    return _make_split(CtrlFilter, mapping)


@_overload.mark(
    """
    CtrlValueSplit(threshold, patch_lower, patch_upper)
    CtrlValueSplit(mapping)

    Split events by controller value.

    The first version splits at a single threshold. The second version allows
    an arbitrary number of (possibly overlapping) value ranges, with *mapping*
    being a dictionary of the form ``{value: patch, ...}`` or
    ``{(lower, upper): patch, ...}``.

    Non-control-change events are discarded.

    """
)
@_arguments.accept(_util.ctrl_limit, _UNIT_TYPES, _UNIT_TYPES)
def CtrlValueSplit(threshold, patch_lower, patch_upper):
    return _make_threshold(CtrlValueFilter(0, threshold),
                           patch_lower, patch_upper)

@_overload.mark
@_arguments.accept({_arguments.nullable(_util.ctrl_range): _UNIT_TYPES})
def CtrlValueSplit(mapping):
    return _make_split(CtrlValueFilter, mapping, unpack=True)


@_arguments.accept({
    _arguments.nullable(_arguments.flatten(_util.program_number, tuple)):
        _UNIT_TYPES
})
def ProgramSplit(mapping):
    """
    ProgramSplit(mapping)

    Split events by program number, with *mapping* being a dictionary of the
    form ``{programs: patch, ...}``.
    Non-program-change events are discarded.
    """
    return _make_split(ProgramFilter, mapping)


@_overload.mark(
    """
    SysExSplit(mapping)
    SysExSplit(manufacturers=...)

    Split events by sysex data or manufacturer id, with *mapping* being a
    dictionary of the form ``{sysex: patch, ...}``, and *manufacturers* being
    a dictionary of the form ``{manufacturer: patch, ...}``
    (cf. :func:`SysExFilter()`).

    Non-sysex events are discarded.
    """
)
@_arguments.accept(dict)
def SysExSplit(mapping):
    return _make_split(SysExFilter, mapping)

@_overload.mark
@_arguments.accept(dict)
def SysExSplit(manufacturers):
    return _make_split(lambda m: SysExFilter(manufacturer=m), manufacturers)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import sys
import tkinter


class AutoScrollbar(tkinter.Scrollbar):
    def set_show_hide(self, show, hide):
        self._show = show
        self._hide = hide

    def set(self, lo, hi):
        if float(lo) <= 0.0 and float(hi) >= 1.0:
            self._hide()
        else:
            self._show()
        tkinter.Scrollbar.set(self, lo, hi)


class LiveThemedFactory(object):
    def __init__(self, color, color_highlight, color_background):
        self.color = color
        self.color_highlight = color_highlight
        self.color_background = color_background

    def Tk(self, **options):
        w = tkinter.Tk()
        w.config(background=self.color_background)
        w.config(**options)
        return w

    def Frame(self, master, **options):
        w = tkinter.Frame(master)
        w.config(background=self.color)
        w.config(**options)
        return w

    def AutoScrollbar(self, master, **options):
        w = AutoScrollbar(master)
        w.config(
            background=self.color,
            activebackground=self.color_highlight,
            troughcolor=self.color_background,
            borderwidth=1,
            relief='flat',
            width=16,
        )
        w.config(**options)
        return w

    def Listbox(self, master, **options):
        w = tkinter.Listbox(master)
        w.config(
            background=self.color_background,
            foreground=self.color,
            selectbackground=self.color_background,
            selectforeground=self.color_highlight,
            selectborderwidth=0,
            borderwidth=0,
        )
        w.config(**options)
        return w

    def Button(self, master, **options):
        w = tkinter.Button(master)
        w.config(
            background=self.color_background,
            foreground=self.color,
            activebackground=self.color_background,
            activeforeground=self.color_highlight,
            borderwidth=0,
            highlightthickness=0,
            relief='flat',
        )
        w.config(**options)
        return w

    def Canvas(self, master, **options):
        w = tkinter.Canvas(master)
        w.config(background=self.color_background)
        w.config(**options)
        return w

class UnthemedFactory(object):
    def Tk(self, **options):
        w = tkinter.Tk()
        w.config(**options)
        return w

    def Frame(self, master, **options):
        return tkinter.Frame(master, **options)

    def AutoScrollbar(self, master, **options):
        return AutoScrollbar(master, **options)

    def Listbox(self, master, **options):
        return tkinter.Listbox(master, **options)

    def Button(self, master, **options):
        return tkinter.Button(master, **options)

    def Canvas(self, master, **options):
        return tkinter.Canvas(master, **options)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

from mididings.units.base import _Unit

import mididings.constants as _constants
import mididings.util as _util
import mididings.overload as _overload
import mididings.unitrepr as _unitrepr


@_unitrepr.accept(_util.event_type, _util.port_number_ref,
                  _util.channel_number_ref, int, int)
def Generator(type, port=_constants.EVENT_PORT,
              channel=_constants.EVENT_CHANNEL, data1=0, data2=0):
    """
    Generator(type, port, channel, data1=0, data2=0)

    Generic generator to change the incoming event's type and data.
    System common and system realtime events can only be created this way.
    """
    return _Unit(_mididings.Generator(
        type,
        _util.actual_ref(port),
        _util.actual_ref(channel),
        data1,
        data2
    ))


@_overload.partial((_constants.EVENT_PORT, _constants.EVENT_CHANNEL))
@_unitrepr.accept(_util.port_number_ref, _util.channel_number_ref,
                  _util.note_number_ref, _util.velocity_value_ref)
def NoteOn(port, channel, note, velocity):
    """
    NoteOn(note, velocity)
    NoteOn(port, channel, note, velocity)

    Create a note-on event, replacing the incoming event.
    """
    return _Unit(_mididings.Generator(
        _constants.NOTEON,
        _util.actual_ref(port),
        _util.actual_ref(channel),
        note,
        velocity
    ))


@_overload.partial((_constants.EVENT_PORT, _constants.EVENT_CHANNEL))
@_unitrepr.accept(_util.port_number_ref, _util.channel_number_ref,
                  _util.note_number_ref, _util.velocity_value_ref)
def NoteOff(port, channel, note, velocity=0):
    """
    NoteOff(note, velocity=0)
    NoteOff(port, channel, note, velocity=0)

    Create a note-off event, replacing the incoming event.
    """
    return _Unit(_mididings.Generator(
        _constants.NOTEOFF,
        _util.actual_ref(port),
        _util.actual_ref(channel),
        note,
        velocity
    ))


@_overload.partial((_constants.EVENT_PORT, _constants.EVENT_CHANNEL))
@_unitrepr.accept(_util.port_number_ref, _util.channel_number_ref,
                  _util.ctrl_number_ref, _util.ctrl_value_ref)
def Ctrl(port, channel, ctrl, value):
    """
    Ctrl(ctrl, value)
    Ctrl(port, channel, ctrl, value)

    Create a control change event, replacing the incoming event.
    """
    return _Unit(_mididings.Generator(
        _constants.CTRL,
        _util.actual_ref(port),
        _util.actual_ref(channel),
        ctrl,
        value
    ))


@_overload.partial((_constants.EVENT_PORT, _constants.EVENT_CHANNEL))
@_unitrepr.accept(_util.port_number_ref, _util.channel_number_ref, int)
def Pitchbend(port, channel, value):
    """
    Pitchbend(value)
    Pitchbend(port, channel, value)

    Create a pitch-bend event, replacing the incoming event.
    """
    return _Unit(_mididings.Generator(
        _constants.PITCHBEND,
        _util.actual_ref(port),
        _util.actual_ref(channel),
        0,
        value
    ))


@_overload.partial((_constants.EVENT_PORT, _constants.EVENT_CHANNEL))
@_unitrepr.accept(_util.port_number_ref, _util.channel_number_ref, int)
def Aftertouch(port, channel, value):
    """
    Aftertouch(value)
    Aftertouch(port, channel, value)

    Create an aftertouch event, replacing the incoming event.
    """
    return _Unit(_mididings.Generator(
        _constants.AFTERTOUCH,
        _util.actual_ref(port),
        _util.actual_ref(channel),
        0,
        value
    ))


@_overload.partial((_constants.EVENT_PORT, _constants.EVENT_CHANNEL))
@_unitrepr.accept(_util.port_number_ref, _util.channel_number_ref,
                  _util.note_number_ref, int)
def PolyAftertouch(port, channel, note, value):
    """
    PolyAftertouch(note, value)
    PolyAftertouch(port, channel, note, value)

    Create a polyphonic aftertouch event, replacing the incoming event.
    """
    return _Unit(_mididings.Generator(
        _constants.POLY_AFTERTOUCH,
        _util.actual_ref(port),
        _util.actual_ref(channel),
        note,
        value
    ))


@_overload.partial((_constants.EVENT_PORT, _constants.EVENT_CHANNEL))
@_unitrepr.accept(_util.port_number_ref, _util.channel_number_ref,
                  _util.program_number_ref)
def Program(port, channel, program):
    """
    Program(program)
    Program(port, channel, program)

    Create a program change event, replacing the incoming event.
    """
    return _Unit(_mididings.Generator(
        _constants.PROGRAM,
        _util.actual_ref(port),
        _util.actual_ref(channel),
        0,
        _util.actual_ref(program)
    ))


@_overload.partial((_constants.EVENT_PORT,))
@_unitrepr.accept(_util.port_number_ref, _util.sysex_data)
def SysEx(port, sysex):
    """
    SysEx(sysex)
    SysEx(port, sysex)

    Create a system exclusive event, replacing the incoming event.
    *sysex* can be a string (binary or ASCII) or a sequence of integers,
    and must include the leading ``F0`` and trailing ``F7`` bytes.
    """
    return _Unit(_mididings.SysExGenerator(
        _util.actual_ref(port),
        sysex,
    ))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

from mididings.units.base import _Unit, Filter, Split, Pass
from mididings.units.splits import VelocitySplit
from mididings.units.generators import NoteOn, NoteOff, PolyAftertouch

import mididings.util as _util
import mididings.misc as _misc
import mididings.overload as _overload
import mididings.constants as _constants
import mididings.arguments as _arguments
import mididings.unitrepr as _unitrepr


@_unitrepr.accept(_util.port_number)
def Port(port):
    """
    Port(port)

    Change the event's port number.
    """
    return _Unit(_mididings.Port(_util.actual(port)))


@_unitrepr.accept(_util.channel_number)
def Channel(channel):
    """
    Channel(channel)

    Change the event's channel number.
    """
    return _Unit(_mididings.Channel(_util.actual(channel)))


@_overload.mark(
    """
    Transpose(offset)
    Transpose(octaves=...)

    Transpose note events by the given number of semitones or octaves.
    """
)
@_unitrepr.accept(int)
def Transpose(offset):
    return _Unit(_mididings.Transpose(offset))

@_overload.mark
@_arguments.accept(int)
def Transpose(octaves):
    return Transpose(octaves * 12)


@_unitrepr.accept(_util.note_number)
def Key(note):
    """
    Key(note)

    Change note events to a fixed note number.
    """
    return _Unit(_mididings.Key(note))


@_overload.mark(
    """
    Velocity(offset)
    Velocity(multiply=...)
    Velocity(fixed=...)
    Velocity(curve=...)
    Velocity(gamma=...)
    Velocity(multiply, offset)

    Change the velocity of note-on events:

    - *offset*: add the given value to the event's velocity.
    - *multiply*: multiply the event's velocity with the given value.
    - *fixed*: set the event's velocity to a fixed value.
    - *curve*: apply an exponential function
      ``f(x) = 127 * (exp(p*x/127)-1) / (exp(p)-1)``.
      Positive values increase velocity, while negative values decrease it.
    - *gamma*: apply a simple power function
      ``f(x) = 127 * (x/127)**(1/p)``.
      Values greater than 1 increase velocity, while values between 0 and 1
      decrease it.
    """
)
@_unitrepr.accept(int)
def Velocity(offset):
    return _Unit(_mididings.Velocity(
                    offset, _mididings.TransformMode.OFFSET))

@_overload.mark
@_unitrepr.accept((float, int))
def Velocity(multiply):
    return _Unit(_mididings.Velocity(
                    multiply, _mididings.TransformMode.MULTIPLY))

@_overload.mark
@_unitrepr.accept(_util.velocity_value)
def Velocity(fixed):
    return _Unit(_mididings.Velocity(
                    fixed, _mididings.TransformMode.FIXED))

@_overload.mark
@_unitrepr.accept((float, int))
def Velocity(gamma):
    return _Unit(_mididings.Velocity(
                    gamma, _mididings.TransformMode.GAMMA))

@_overload.mark
@_unitrepr.accept((float, int))
def Velocity(curve):
    return _Unit(_mididings.Velocity(
                    curve, _mididings.TransformMode.CURVE))

@_overload.mark
@_unitrepr.accept((float, int), int)
def Velocity(multiply, offset):
    return Velocity(multiply=multiply) >> Velocity(offset=offset)


@_overload.mark(
    """
    VelocitySlope(notes, offset)
    VelocitySlope(notes, multiply=...)
    VelocitySlope(notes, fixed=...)
    VelocitySlope(notes, curve=...)
    VelocitySlope(notes, gamma=...)
    VelocitySlope(notes, multiply, offset)

    Change the velocity of note-on events, applying a linear slope between
    different notes. This can be thought of as a :func:`~.Velocity()` unit
    with different parameters for different note ranges, and is useful for
    example to fade-in a sound over a region of the keyboard.

    Both parameters must be sequences of the same length, with notes in
    ascending order, and one velocity parameter corresponding to each note.
    """
)
@_unitrepr.accept([_util.note_limit], [int])
def VelocitySlope(notes, offset):
    _check_velocity_slope(notes, offset)
    return _Unit(_mididings.VelocitySlope(
                    notes, offset, _mididings.TransformMode.OFFSET))

@_overload.mark
@_unitrepr.accept([_util.note_limit], [(float, int)])
def VelocitySlope(notes, multiply):
    _check_velocity_slope(notes, multiply)
    return _Unit(_mididings.VelocitySlope(
                    notes, multiply, _mididings.TransformMode.MULTIPLY))

@_overload.mark
@_unitrepr.accept([_util.note_limit], [_util.velocity_value])
def VelocitySlope(notes, fixed):
    _check_velocity_slope(notes, fixed)
    return _Unit(_mididings.VelocitySlope(
                    notes, fixed, _mididings.TransformMode.FIXED))

@_overload.mark
@_unitrepr.accept([_util.note_limit], [(float, int)])
def VelocitySlope(notes, gamma):
    _check_velocity_slope(notes, gamma)
    return _Unit(_mididings.VelocitySlope(
                    notes, gamma, _mididings.TransformMode.GAMMA))

@_overload.mark
@_unitrepr.accept([_util.note_limit], [(float, int)])
def VelocitySlope(notes, curve):
    _check_velocity_slope(notes, curve)
    return _Unit(_mididings.VelocitySlope(
                    notes, curve, _mididings.TransformMode.CURVE))

@_overload.mark
@_unitrepr.accept([_util.note_limit], [(float, int)], [int])
def VelocitySlope(notes, multiply, offset):
    return (VelocitySlope(notes, multiply=multiply)
            >> VelocitySlope(notes, offset=offset))


def _check_velocity_slope(notes, params):
    message = None
    if len(notes) != len(params):
        message = ("invalid parameters to VelocitySlope(): notes and velocity"
                   "values must be sequences of the same length")
    elif len(notes) < 2:
        message = ("invalid parameters to VelocitySlope(): need at least"
                   "two notes")
    elif sorted(notes) != list(notes):
        message = ("invalid parameters to VelocitySlope(): notes must be in"
                   "ascending order")

    if message is not None:
        raise ValueError(message)


@_overload.mark(
    """
    VelocityLimit(min, max)
    VelocityLimit(min)
    VelocityLimit(max=...)

    Limit velocities of note-on events to the given range.
    """
)
@_arguments.accept(_util.velocity_limit, _util.velocity_limit)
def VelocityLimit(min, max):
    return Filter(_constants.NOTE) % VelocitySplit({
        (0, min):   Velocity(fixed=min),
        (min, max): Pass(),
        (max, 0):   Velocity(fixed=max),
    })

@_overload.mark
@_arguments.accept(_util.velocity_limit)
def VelocityLimit(max):
    return Filter(_constants.NOTE) % VelocitySplit({
        (0, max):   Pass(),
        (max, 0):   Velocity(fixed=max),
    })

@_overload.mark
@_arguments.accept(_util.velocity_limit)
def VelocityLimit(min):
    return Filter(_constants.NOTE) % VelocitySplit({
        (0, min):   Velocity(fixed=min),
        (min, 0):   Pass(),
    })


@_unitrepr.accept(_util.ctrl_number, _util.ctrl_number)
def CtrlMap(ctrl_in, ctrl_out):
    """
    CtrlMap(ctrl_in, ctrl_out)

    Convert controller *ctrl_in* to *ctrl_out*, i.e. change the event's
    control change number.
    """
    return _Unit(_mididings.CtrlMap(ctrl_in, ctrl_out))


@_unitrepr.accept(_util.ctrl_number, int, int, int, int)
def CtrlRange(ctrl, min, max, in_min=0, in_max=127):
    """
    CtrlRange(ctrl, min, max, in_min=0, in_max=127)

    Linearly map control change values for controller *ctrl* from the interval
    [*in_min*, *in_max*] to the interval [*min*, *max*].
    Any input value less than or equal to *in_min* results in an output value
    of *min*. Likewise, any value of *in_max* or greater results in an output
    value of *max*.
    """
    if in_min > in_max:
        # swap ranges so that in_min is less than in_max
        in_min, in_max = in_max, in_min
        min, max = max, min
    return _Unit(_mididings.CtrlRange(ctrl, min, max, in_min, in_max))


@_overload.mark(
    """
    CtrlCurve(ctrl, gamma)
    CtrlCurve(ctrl, curve=...)
    CtrlCurve(ctrl, offset=...)
    CtrlCurve(ctrl, multiply=...)
    CtrlCurve(ctrl, multiply, offset)

    Transform control change values. See :func:`~.Velocity()` for a
    description of the parameters.
    """
)
@_unitrepr.accept(_util.ctrl_number, (float, int))
def CtrlCurve(ctrl, gamma):
    return _Unit(_mididings.CtrlCurve(
                        ctrl, gamma, _mididings.TransformMode.GAMMA))

@_overload.mark
@_unitrepr.accept(_util.ctrl_number, (float, int))
def CtrlCurve(ctrl, curve):
    return _Unit(_mididings.CtrlCurve(
                        ctrl, curve, _mididings.TransformMode.CURVE))

@_overload.mark
@_unitrepr.accept(_util.ctrl_number, int)
def CtrlCurve(ctrl, offset):
    return _Unit(_mididings.CtrlCurve(
                        ctrl, offset, _mididings.TransformMode.OFFSET))

@_overload.mark
@_unitrepr.accept(_util.ctrl_number, (float, int))
def CtrlCurve(ctrl, multiply):
    return _Unit(_mididings.CtrlCurve(
                        ctrl, multiply, _mididings.TransformMode.MULTIPLY))

@_overload.mark
@_unitrepr.accept(_util.ctrl_number, (float, int), int)
def CtrlCurve(ctrl, multiply, offset):
    return (CtrlCurve(ctrl, multiply=multiply)
            >> CtrlCurve(ctrl, offset=offset))


@_overload.mark(
    """
    PitchbendRange(min, max, in_min=-8192, in_max=8191)
    PitchbendRange(down, up, range=...)

    Change the pitchbend range to values between *min* and *max*, or to the
    given number of semitones *down* and *up*. The latter requires the tone
    generator's pitchbend range to be specified as *range*.
    """
)
@_unitrepr.accept(int, int, int, int)
def PitchbendRange(min, max, in_min=-8192, in_max=8191):
    return _Unit(_mididings.PitchbendRange(min, max, in_min, in_max))

@_overload.mark
@_unitrepr.accept(int, int, int)
def PitchbendRange(down, up, range):
    return PitchbendRange(int(float(down)/range*8192),
                          int(float(up)/range*8191))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings.setup as _setup
import mididings.engine as _engine

import sys as _sys


class MemorizeScene(object):
    """
    Saves the currently selected scene number to a file when terminating
    mididings, and restores it at the next startup.

    :param memo_file:
        the path of the file to be used to store the scene number.
    """
    def __init__(self, memo_file):
        self.memo_file = memo_file

    def on_start(self):
        try:
            f = open(self.memo_file)
            s = f.read()
            try:
                # single scene number
                _setup.config(initial_scene=int(s))
            except ValueError:
                try:
                    # scene and subscene number
                    _setup.config(initial_scene=tuple(map(int, s.split())))
                except ValueError:
                    pass
        except IOError:
            # couldn't open memo file, might not be an error
            pass

    def on_exit(self):
        try:
            f = open(self.memo_file, 'w')
            f.write("%d %d\n" % (_engine.current_scene(),
                                 _engine.current_subscene()))
        except IOError as ex:
            print(f"couldn't store current scene:\n{ex}")



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings import Call as _Call
import mididings.engine as _engine


def Restart():
    """
    Call :func:`.engine.restart()`.
    """
    return _Call(lambda ev: _engine.restart())

def Quit():
    """
    Call :func:`.engine.quit()`.
    """
    return _Call(lambda ev: _engine.quit())



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings.extra.harmonizer import *
from mididings.extra.suppress_pc import *
from mididings.extra.pedal_noteoff import *
from mididings.extra.polyphony import *
from mididings.extra.key_color import *
from mididings.extra.memorize_scene import *
from mididings.extra.latch import *
from mididings.extra.panic import *
from mididings.extra.floating_split import *
from mididings.extra.voices import *
from mididings.extra.engine import *
from mididings.extra.ctrl_sysex import *


import mididings.misc as _misc
__all__ = _misc.prune_globals(globals())



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m
import mididings.event as _event
from mididings.extra.per_channel import PerChannel as _PerChannel


class _LimitPolyphony(object):
    def __init__(self, max_polyphony, remove_oldest):
        self.max_polyphony = max_polyphony
        self.remove_oldest = remove_oldest
        self.notes = []

    def __call__(self, ev):
        if ev.type == _m.NOTEON:
            if len(self.notes) < self.max_polyphony:
                # polyphony not exceeded, allow note
                self.notes.append(ev.note)
                return ev
            else:
                if self.remove_oldest:
                    # allow note, but send note-off for oldest first
                    noteoff = _event.NoteOffEvent(ev.port, ev.channel,
                                                  self.notes[0], 0)
                    self.notes = self.notes[1:]
                    self.notes.append(ev.note)
                    return [noteoff, ev]
                else:
                    # discard note
                    return None

        elif ev.type == _m.NOTEOFF:
            if ev.note in self.notes:
                self.notes.remove(ev.note)
                return ev
            else:
                return None


class _MakeMonophonic(object):
    def __init__(self):
        self.notes = []

    def __call__(self, ev):
        if ev.type == _m.NOTEON:
            if len(self.notes):
                # send note off for previous note, and note on for current note
                noteoff = _event.NoteOffEvent(ev.port, ev.channel,
                                              self.notes[-1][0], 0)
                self.notes.append((ev.note, ev.velocity))
                return [noteoff, ev]
            else:
                # send note on for current note
                self.notes.append((ev.note, ev.velocity))
                return ev

        elif ev.type == _m.NOTEOFF:
            if len(self.notes) and ev.note == self.notes[-1][0]:
                # note off for currently sounding note
                if len(self.notes) == 1:
                    # only one note left, send note off
                    r = ev
                else:
                    # send note off, and retrigger previous note
                    noteon = _event.NoteOnEvent(ev.port, ev.channel,
                                        self.notes[-2][0], self.notes[-2][1])
                    r = [ev, noteon]
            else:
                # note isn't sounding, discard note off
                r = None

            # remove released note from list
            self.notes = [x for x in self.notes if x[0] != ev.note]

            return r



def LimitPolyphony(max_polyphony, remove_oldest=True):
    """
    Limit the "MIDI polyphony".

    :param max_polyphony:
        The maximum number of simultaneous notes.

    :param remove_oldest:
        If true, the oldest notes will be stopped when the maximum polyphony
        is exceeded.
        If false, no new notes are accepted while *max_polyphony* notes
        are already held.

    Note that the actual polyphony of a connected synthesizer can still be
    higher than the limit set here, e.g. due to a long release phase.
    """
    return (_m.Filter(_m.NOTE) %
        _m.Process(_PerChannel(
            lambda: _LimitPolyphony(max_polyphony, remove_oldest))))


def MakeMonophonic():
    """
    Make the MIDI signal monophonic, i.e. only one note can be played at
    any given time.
    When one note is released while another is still held (but silent),
    the previous one will be retriggered.
    """
    return (_m.Filter(_m.NOTE) %
        _m.Process(_PerChannel(_MakeMonophonic)))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import pyliblo3 as liblo


class LiveOSC(liblo.ServerThread):
    def __init__(self, dings, control_port, listen_port):
        liblo.ServerThread.__init__(self, listen_port)

        self.dings = dings
        self.control_port = control_port

        self._scenes = {}

    def query(self):
        self.send(self.control_port, '/mididings/query')

    def switch_scene(self, n):
        self.send(self.control_port, '/mididings/switch_scene', n)

    def switch_subscene(self, n):
        self.send(self.control_port, '/mididings/switch_subscene', n)

    def prev_scene(self):
        self.send(self.control_port, '/mididings/prev_scene')

    def next_scene(self):
        self.send(self.control_port, '/mididings/next_scene')

    def prev_subscene(self):
        self.send(self.control_port, '/mididings/prev_subscene')

    def next_subscene(self):
        self.send(self.control_port, '/mididings/next_subscene')

    def panic(self):
        self.send(self.control_port, '/mididings/panic')

    def quit(self):
        self.send(self.control_port, '/mididings/quit')

    def restart(self):
        self.send(self.control_port, '/mididings/restart')

    @liblo.make_method('/mididings/data_offset', 'i')
    def data_offset_cb(self, path, args):
        self.dings.set_data_offset(args[0])

    @liblo.make_method('/mididings/begin_scenes', '')
    def begin_scenes_cb(self, path, args):
        self._scenes = {}

    @liblo.make_method('/mididings/add_scene', None)
    def add_scene_cb(self, path, args):
        number, name = args[:2]
        subscenes = args[2:]
        self._scenes[number] = (name, subscenes)

    @liblo.make_method('/mididings/end_scenes', '')
    def end_scenes_cb(self, path, args):
        self.dings.set_scenes(self._scenes)

    @liblo.make_method('/mididings/current_scene', 'ii')
    def current_scene_cb(self, path, args):
        self.dings.set_current_scene(args[0], args[1])

    @liblo.make_method('/mididings/exit', '')
    def on_exit_cb(self, path, args):
        self.dings.on_exit()

    @liblo.make_method('/mididings/start', '')
    def on_start_cb(self, path, args):
        self.dings.on_start()



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings.live.widgets import LiveThemedFactory, UnthemedFactory
from mididings.live.osc_control import LiveOSC

import sys


class LiveDings(object):
    def __init__(self, options):
        self.options = options

        # start OSC server
        self.osc = LiveOSC(self, self.options.control_port,
                                 self.options.listen_port)
        self.osc.start()

        if self.options.themed:
            widget_factory = LiveThemedFactory(self.options.color,
                                               self.options.color_highlight,
                                               self.options.color_background)
        else:
            widget_factory = UnthemedFactory()

        # create the main window
        self.win = widget_factory.Tk(padx=8, pady=8)
        self.win.minsize(480, 120)
        self.win.geometry(f"{self.options.width}x{self.options.height}")

        if self.options.name:
            self.win.title(f"livedings - {self.options.name}")
        else:
            self.win.title('livedings')

        # track window resizing
        self.win.bind('<Configure>',
                      lambda event: self.win.after_idle(self.update, True))

        # configure the grid
        self.win.grid_rowconfigure(0, weight=1)
        self.win.grid_columnconfigure(1, minsize=self.options.list_width,
                                      weight=0)
        for n in range(3, 8):
            self.win.grid_columnconfigure(n, weight=1, minsize=64)

        # create listbox
        self.listbox = widget_factory.Listbox(
                            self.win, font=self.options.list_font,
                            selectmode='single', activestyle='none',
                            highlightthickness=0)
        self.listbox.grid(column=1, row=0, rowspan=2, sticky='nsew', padx=8)
        self.listbox.bind('<ButtonRelease-1>',
                          lambda event: self.on_select_scene())

        # create scrollbar for listbox. will be attached to the grid only
        # when necessary
        self.scrollbar = widget_factory.AutoScrollbar(self.win,
                                                      orient='vertical')
        self.scrollbar.set_show_hide(
            lambda: self.scrollbar.grid(column=0, row=0,
                                        rowspan=2, sticky='ns'),
            lambda: self.scrollbar.grid_forget()
        )
        self.scrollbar.config(command=self.listbox.yview)
        self.listbox.config(yscrollcommand=self.scrollbar.set)

        # create separator
        separator = widget_factory.Frame(self.win, width=2)
        separator.grid(column=2, row=0, rowspan=2, sticky='ns')

        # create canvas
        self.canvas = widget_factory.Canvas(self.win, highlightthickness=0)
        if self.options.color_background is not None:
            self.canvas.config(background=self.options.color_background)
        self.canvas.grid(column=3, columnspan=5, row=0, sticky='nsew')
        self.canvas.bind('<Button-1>', self.on_button_press)
        self.canvas.bind('<ButtonRelease-1>', self.on_button_release)

        # create buttons
        try:
            button_size = int(int(self.options.font.split(' ')[1]) / 1.5)
        except IndexError:
            button_size = 20
        button_font = f"Sans {button_size} bold"

        self.btn_prev_scene = widget_factory.Button(
                                    self.win, text=chr(0x25c0)*2,
                                    width=64, font=button_font,
                                    command=self.osc.prev_scene)
        self.btn_prev_scene.grid(column=3, row=1, padx=8)

        self.btn_next_scene = widget_factory.Button(
                                    self.win, text=chr(0x25b6)*2,
                                    width=64, font=button_font,
                                    command=self.osc.next_scene)
        self.btn_next_scene.grid(column=4, row=1, padx=8)

        self.btn_prev_subscene = widget_factory.Button(
                                    self.win, text=chr(0x25c0),
                                    width=64, font=button_font,
                                    command=self.osc.prev_subscene)
        self.btn_prev_subscene.grid(column=5, row=1, padx=8)

        self.btn_next_subscene = widget_factory.Button(
                                    self.win, text=chr(0x25b6),
                                    width=64, font=button_font,
                                    command=self.osc.next_subscene)
        self.btn_next_subscene.grid(column=6, row=1, padx=8)

        self.btn_panic = widget_factory.Button(
                                    self.win, text="!",
                                    width=64, font=button_font,
                                    command=self.osc.panic)
        self.btn_panic.grid(column=7, row=1, padx=8)

        # attempt to calculate the height of one line in the current font.
        # this is crazy...
        try:
            self.line_height = int(self.options.font.split(' ')[1]) * 2
        except Exception:
            # whatever...
            self.line_height = 72

        # some keybindings
        self.win.bind('<Left>', lambda event: self.osc.prev_subscene())
        self.win.bind('<Right>', lambda event: self.osc.next_subscene())
        self.win.bind('<Up>', lambda event: self.osc.prev_scene())
        self.win.bind('<Down>', lambda event: self.osc.next_scene())
        self.win.bind('<Escape>', lambda event: self.osc.panic())
        # prevent left/right keys from scrolling the listbox
        self.win.bind_class('Listbox', "<Left>", lambda event: None)
        self.win.bind_class('Listbox', "<Right>", lambda event: None)

        # get config from mididings
        self.osc.query()

        self._ready = False

        # window dimensions
        self._width = 0
        self._height = 0
        # mouse click position
        self._click_x = 0
        self._click_y = 0

    def on_select_scene(self):
        cursel = self.listbox.curselection()
        if cursel:
            self.osc.switch_scene(sorted(self.scenes.keys())[int(cursel[0])])

    def on_button_press(self, event):
        self._click_x = event.x
        self._click_y = event.y

    def on_button_release(self, event):
        if (self._ready and
            self._click_y > 8 + 3 * self.line_height and
            self._click_y < 8 + (len(self.scenes[self.current_scene][1])+3) *
                                    self.line_height):
            n = (self._click_y - (8 + 3 * self.line_height)) / self.line_height
            self.osc.switch_subscene(n + self.data_offset)

    def update(self, resize=False):
        width = self.canvas.winfo_width()
        height = self.canvas.winfo_height()

        # check if the window size really changed
        if resize and width == self._width and height == self._height:
            return

        self.draw_canvas(width, height)
        self._width = width
        self._height = height

    def set_data_offset(self, data_offset):
        self.data_offset = data_offset

    def set_scenes(self, scenes):
        self.scenes = scenes
        self.update_scenes()
        self._ready = True

    def set_current_scene(self, scene, subscene):
        self.current_scene = scene
        self.current_subscene = subscene
        self.listbox.selection_clear(0, 'end')
        self.listbox.selection_set(sorted(self.scenes.keys()).index(scene))
        self.update()

    def on_exit(self):
        pass

    def on_start(self):
        pass

    def update_scenes(self):
        self.listbox.delete(0, 'end')
        for n in sorted(self.scenes.keys()):
            name = self.scenes[n][0]
            if name:
                self.listbox.insert('end', f"{int(n)}: {name}")
            else:
                self.listbox.insert('end', f"{int(n)}")

    def draw_canvas(self, width, height):
        if not self._ready or not len(self.scenes):
            return

        scene = self.current_scene
        subscene = self.current_subscene
        scene_name, subscene_names = self.scenes[scene]
        has_subscenes = bool(len(subscene_names))

        if not scene_name:
            scene_name = "(unnamed)"

        self.canvas.delete('all')

        # draw scene number
        self.canvas.create_text(
            24,
            8,
            text=((f"{int(scene)}.{int(subscene)}")
                    if has_subscenes else str(scene)),
            fill=self.options.color,
            font=self.options.font,
            anchor='nw'
        )

        # draw scene name
        self.canvas.create_text(
            width / 2 + 24,
            8 + 1.5 * self.line_height,
            text=scene_name,
            fill=self.options.color_highlight,
            font=self.options.font,
            anchor='n'
        )

        # draw subscenes
        for n, s in enumerate(subscene_names):
            self.canvas.create_text(
                width / 2 + 24,
                8 + (n+3) * self.line_height,
                text=s if s else "(unnamed)",
                fill=(self.options.color_highlight
                        if n + self.data_offset == subscene
                            else self.options.color),
                font=self.options.font,
                anchor='n'
            )

        # draw indicator
        self.canvas.create_text(
            16,
            8 + 1.5 * self.line_height if not has_subscenes
                else 8 + (subscene - self.data_offset + 3) * self.line_height,
            text=chr(0x25b6),
            fill=self.options.color_highlight,
            font=self.options.font,
            anchor='nw'
        )

    def run(self):
        self.win.mainloop()



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings.engine as _engine

import sys as _sys
import os as _os

import pyinotify as _pyinotify


class AutoRestart(object):
    """
    Automatically restarts mididings when the script changes.

    This restarts the entire mididings script, so MIDI processing is
    interrupted, and mididings does not take care of reestablishing any
    ALSA/JACK connections that were made manually.
    If the new script contains errors that prevent it from running,
    mididings exits and needs to be restarted manually once the errors
    are fixed.

    :param modules:
        If true, all imported local Python modules are monitored for
        changes as well.

    :param filenames:
        a list of additional files to be monitored.
    """
    def __init__(self, modules=True, filenames=[]):
        self.modules = modules
        self.filenames = filenames

    def on_start(self):
        self.wm = _pyinotify.WatchManager()
        self.notifier = _pyinotify.ThreadedNotifier(self.wm)

        if self.modules:
            # find the name of the main script being executed
            if '__mididings_main__' in _sys.modules:
                main_file = _sys.modules['__mididings_main__'].__file__
            elif hasattr(_sys.modules['__main__'], '__file__'):
                main_file = _sys.modules['__main__'].__file__
            else:
                main_file = None

            if main_file:
                base_dir = _os.path.dirname(_os.path.abspath(main_file))

                # add watches for imported modules
                for m in _sys.modules.values():
                    # builtin modules don't have a __file__ attribute
                    if hasattr(m, '__file__'):
                        f = _os.path.abspath(m.__file__)
                        # only watch file if it's in the same directory as the
                        # main script
                        if f.startswith(base_dir):
                            self.wm.add_watch(f, _pyinotify.IN_MODIFY,
                                              self._process_IN_MODIFY)

        # add watches for additional files
        for f in self.filenames:
            self.wm.add_watch(f, _pyinotify.IN_MODIFY,
                              self._process_IN_MODIFY)

        self.notifier.start()

    def on_exit(self):
        self.notifier.stop()

    def _process_IN_MODIFY(self, event):
        print(f"file '{event.pathname}' changed, restarting...")
        _engine.restart()



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings import Call as _Call
import mididings.engine as _engine
import mididings.setup as _setup
import mididings.util as _util
import mididings.misc as _misc

import mididings.extra.panic as _panic

import pyliblo3 as _liblo


class OSCInterface(object):
    """
    Allows controlling mididings via OSC.

    :param port:
        the OSC port to listen on.

    :param notify_ports:
        a list of OSC ports to notify when the current scene changes.

    These messages are currently understood:

    - **/mididings/switch_scene ,i**: switch to the given scene number.
    - **/mididings/switch_subscene ,i**: switch to the given subscene number.
    - **/mididings/prev_scene**: switch to the previous scene.
    - **/mididings/next_scene**: switch to the next scene.
    - **/mididings/prev_subscene**: switch to the previous subscene.
    - **/mididings/next_subscene**: switch to the next subscene.
    - **/mididings/panic**: send all-notes-off on all channels and on all
      output ports.
    - **/mididings/quit**: terminate mididings.
    - **/mididings/restart**: restart mididings.
    """
    def __init__(self, port=56418, notify_ports=[56419]):
        self.port = port
        if _misc.issequence(notify_ports):
            self.notify_ports = notify_ports
        else:
            self.notify_ports = [notify_ports]

    def on_start(self):
        if self.port is not None:
            self.server = _liblo.ServerThread(self.port)
            self.server.register_methods(self)
            self.server.start()

        self.send_config(True)

    def on_exit(self):
        if self.port is not None:
            for p in self.notify_ports:
                _liblo.send(p, '/mididings/exit')
            self.server.stop()
            del self.server

    def on_switch_scene(self, scene, subscene):
        for p in self.notify_ports:
            _liblo.send(p, '/mididings/current_scene', scene, subscene)

    def send_config(self, start=False):
        for p in self.notify_ports:
            if start: _liblo.send(p, '/mididings/start')
            # send data offset
            _liblo.send(p, '/mididings/data_offset',
                        _setup.get_config('data_offset'))

            # send list of scenes
            _liblo.send(p, '/mididings/begin_scenes')
            s = _engine.scenes()
            for n in sorted(s.keys()):
                _liblo.send(p, '/mididings/add_scene', n, s[n][0], *s[n][1])
            _liblo.send(p, '/mididings/end_scenes')

    @_liblo.make_method('/mididings/query', '')
    def query_cb(self, path, args):
        self.send_config()
        for p in self.notify_ports:
            _liblo.send(p, '/mididings/current_scene',
                        _engine.current_scene(), _engine.current_subscene())

    @_liblo.make_method('/mididings/switch_scene', 'i')
    @_liblo.make_method('/mididings/switch_scene', 'ii')
    def switch_scene_cb(self, path, args):
        _engine.switch_scene(*args)

    @_liblo.make_method('/mididings/switch_subscene', 'i')
    def switch_subscene_cb(self, path, args):
        _engine.switch_subscene(*args)

    @_liblo.make_method('/mididings/prev_scene', '')
    def prev_scene_cb(self, path, args):
        s = sorted(_engine.scenes().keys())
        n = s.index(_engine.current_scene()) - 1
        if n >= 0:
            _engine.switch_scene(s[n])

    @_liblo.make_method('/mididings/next_scene', '')
    def next_scene_cb(self, path, args):
        s = sorted(_engine.scenes().keys())
        n = s.index(_engine.current_scene()) + 1
        if n < len(s):
            _engine.switch_scene(s[n])

    @_liblo.make_method('/mididings/prev_subscene', '')
    @_liblo.make_method('/mididings/prev_subscene', 'i')
    def prev_subscene_cb(self, path, args):
        s = _engine.scenes()[_engine.current_scene()]
        n = _util.actual(_engine.current_subscene()) - 1
        if len(s[1]) and len(args) and args[0]:
            n %= len(s[1])
        if n >= 0:
            _engine.switch_subscene(_util.offset(n))

    @_liblo.make_method('/mididings/next_subscene', '')
    @_liblo.make_method('/mididings/next_subscene', 'i')
    def next_subscene_cb(self, path, args):
        s = _engine.scenes()[_engine.current_scene()]
        n = _util.actual(_engine.current_subscene()) + 1
        if len(s[1]) and len(args) and args[0]:
            n %= len(s[1])
        if n < len(s[1]):
            _engine.switch_subscene(_util.offset(n))

    @_liblo.make_method('/mididings/panic', '')
    def panic_cb(self, path, args):
        _panic._panic_bypass()

    @_liblo.make_method('/mididings/restart', '')
    def restart_cb(self, path, args):
        _engine.restart()

    @_liblo.make_method('/mididings/quit', '')
    def quit_cb(self, path, args):
        _engine.quit()



class _SendOSC(object):
    def __init__(self, target, path, args):
        self.target = target
        self.path = path
        self.args = args

    def __call__(self, ev):
        args = ((x(ev) if hasattr(x, '__call__') else x) for x in self.args)
        _liblo.send(self.target, self.path, *args)


def SendOSC(target, path, *args):
    """
    Send an OSC message.
    Parameters are the same as for :func:`liblo.send()`.
    Additionally, instead of a specific value, each data argument may also
    be a Python function that takes a single :class:`~.MidiEvent` parameter,
    and returns the value to be sent.
    """
    return _Call(_SendOSC(target, path, args))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m
import mididings.util as _util
import mididings.misc as _misc

import itertools as _itertools
from operator import itemgetter as _itemgetter


_MAJOR_SCALE = [0, 2, 4, 5, 7, 9, 11]
_HARMONIC_MINOR_SCALE = [0, 2, 3, 5, 7, 8, 11]

_MODES = ['ionian', 'dorian', 'phrygian', 'lydian',
          'mixolydian', 'aeolian', 'locrian']

_INTERVALS = ['unison', 'second', 'third', 'fourth', 'fifth',
              'sixth', 'seventh', 'octave', 'ninth', 'tenth',
              'eleventh', 'twelfth', 'thirteenth']


class _Harmonizer(object):
    def __init__(self, tonic, scale, interval, non_harmonic):
        self.tonic = tonic
        self.lookup = {}

        # for each note in the scale, calculate the transpose interval
        for x in scale:
            l = len(scale)
            i = scale.index(x) + interval

            hx = scale[i % l] + ((i // l) * 12)

            self.lookup[x] = hx - x

        r = list(range(12))
        if non_harmonic == 'above':
            # reverse order to make sure higher notes are already calculated
            r.reverse()
        # fill the gaps (notes not in the scale)
        for x in r:
            if x not in scale:
                if non_harmonic == 'below':
                    self.lookup[x] = self.lookup[x-1]
                elif non_harmonic == 'above':
                    self.lookup[x] = self.lookup[(x+1) % 12]
                elif non_harmonic == 'same':
                    self.lookup[x] = 0
                else: # skip
                    self.lookup[x] = None

    def note_offset(self, note):
        n = (note - self.tonic) % 12
        return self.lookup[n]

    def __call__(self, ev):
        off = self.note_offset(ev.note)
        if off is not None:
            ev.note += off
            return True
        else:
            return False


def Harmonize(tonic, scale, interval, non_harmonic='below'):
    """
     A diatonic harmonizer.

    :param tonic:
        The tonic of the scale, as a note name.

    :param scale:
        The type/mode, of the scale, one of:
        ``'major'``, ``'minor'``, ``'minor_harmonic'``, ``'ionian'``,
        ``'dorian'``, ``'phrygian'``, ``'lydian'``, ``'mixolydian'``,
        ``'aeolian'``, ``'locrian'``.

    :param interval:
        The number of steps to transpose the notes by
        (as an integer), or one of these interval names:
        ``'unison'``, ``'second'``, ``'third'``, ``'fourth'``, ``'fifth'``,
        ``'sixth'``, ``'seventh'``, ``'octave'``, ``'ninth'``, ``'tenth'``,
        ``'eleventh'``, ``'twelfth'``, ``'thirteenth'``.

        It is also possible to pass a list of intervals, to create
        multiple harmonized voices.

    :param non_harmonic: What to do with out-of-scale notes:

        - ``'below'``: Transpose by the same interval as the next on-scale
        - ``'above'``: Transpose by the same interval as the next on-scale
        - ``'skip'``: Ignore note.
        - ``'same'``: Output note as is, without transposing it.
    """
    t = _util.tonic_note_number(tonic)

    if _misc.issequence(scale):
        shift = 0
    elif isinstance(scale, str):
        if scale == 'major':
            scale = _MAJOR_SCALE
            shift = 0
        elif scale == 'minor':
            scale = _MAJOR_SCALE
            shift = 5
        elif scale == 'minor_harmonic':
            scale = _HARMONIC_MINOR_SCALE
            shift = 0
        elif scale in _MODES:
            shift = _MODES.index(scale)
            scale = _MAJOR_SCALE

    # shift scale to the correct mode
    s = ([x - scale[shift] for x in scale[shift:]] +
         [x + 12 - scale[shift] for x in scale[:shift]])

    if not _misc.issequence(interval):
        interval = [interval]

    # convert all interval names to numbers
    iv = [(_INTERVALS.index(x) if x in _INTERVALS else x) for x in interval]

    # python version:
#    f = [ _m.Process(_Harmonizer(t, s, i, non_harmonic)) for i in iv ]

    # pure mididings version:
    f = []
    for i in iv:
        h = _Harmonizer(t, s, i, non_harmonic)
        # get offset for each key
        offsets = [(x, h.note_offset(x)) for x in range(128)]
        # group by offset
        groups = _itertools.groupby(sorted(offsets, key=_itemgetter(1)),
                                    key=_itemgetter(1))

        # create one KeyFilter()/Transpose() pair for each offset
        for off, keys in groups:
            if off is not None:
                f.append(_m.KeyFilter(notes=[k[0] for k in keys]) >>
                         _m.Transpose(off))

    return _m.Filter(_m.NOTE) % f



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m
from mididings.extra.per_channel import PerChannel as _PerChannel


class _SuppressPC(object):
    def __init__(self):
        self.current = None
    def __call__(self, ev):
        if ev.program == self.current:
            return None
        else:
            self.current = ev.program
            return ev


def SuppressPC():
    """
    Filter out program changes if the same program has already
    been selected on the same port/channel.
    """
    return (_m.Filter(_m.PROGRAM) %
        _m.Process(_PerChannel(_SuppressPC)))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m


_BLACK = [1, 3, 6, 8, 10]
_BLACK_KEYS = [_n for _n in range(128) if _n%12 in _BLACK]
_WHITE_KEYS = [_n for _n in range(128) if _n%12 not in _BLACK]


def KeyColorFilter(color):
    """
    Filter notes by key color.

    :param color:
        the key color, either ``'black'`` or ``'white'``.
    """
    if color == 'black':
        return _m.KeyFilter(notes=_BLACK_KEYS)
    elif color == 'white':
        return _m.KeyFilter(notes=_WHITE_KEYS)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

from mididings.units.base import _Unit

import mididings.event as _event
import mididings.overload as _overload
import mididings.unitrepr as _unitrepr
import mididings.misc as _misc
from mididings.setup import get_config as _get_config

import sys as _sys
import _thread
import subprocess as _subprocess
import types as _types
import copy as _copy
import collections as _collections
import collections.abc as _collections_abc
import inspect as _inspect


class _CallBase(_Unit):
    def __init__(self, function, asyncFlag, cont):
        def do_call(ev):
            # add additional properties that don't exist on the C++ side
            ev.__class__ = _event.MidiEvent

            # call the function
            ret = function(ev)

            if ret is None or asyncFlag:
                return None
            elif isinstance(ret, _types.GeneratorType):
                # function is a generator, build list
                ret = list(ret)
            elif not _misc.issequence(ret):
                ret = [ret]

            for ev in ret:
                ev._finalize()
            return ret

        _Unit.__init__(self, _mididings.Call(do_call, asyncFlag, cont))


class _CallThread(_CallBase):
    def __init__(self, function):
        def do_thread(ev):
            # need to make a copy of the event. the underlying C++ object will
            # become invalid once this function returns
            ev_copy = _copy.copy(ev)
            _thread.start_new_thread(function, (ev_copy,))

        _CallBase.__init__(self, do_thread, True, False)


class _System(_CallBase):
    def __init__(self, command):
        def do_system(ev):
            args = command(ev) if hasattr(command, '__call__') else command
            _subprocess.Popen(args, shell=True)

        _CallBase.__init__(self, do_system, True, False)


def _call_partial(function, args, kwargs, require_event=False):
    """
    Add args and kwargs when calling function (similar to functools.partial).
    Also, allow omission of the first argument (event) if require_event
    is False.
    """
    try:
        argspec = _inspect.getfullargspec(function)
        ismethod = _inspect.ismethod(function)
    except TypeError:
        # support callable objects
        argspec = _inspect.getfullargspec(function.__call__)
        ismethod = _inspect.ismethod(function.__call__)

    if (not require_event and argspec[1] == None
                and len(argspec[0]) - int(ismethod) == 0):
        # omit event argument when function has no positional arguments
        if len(kwargs):
            return lambda ev: function(**kwargs)
        else:
            return lambda ev: function()
    if len(args) or len(kwargs):
        # return a function with args and kwargs applied
        return lambda ev: function(ev, *args, **kwargs)
    else:
        # no additional arguments, no wrapper needed
        return function


@_unitrepr.accept(_collections_abc.Callable, None, kwargs={ None: None })
def Process(function, *args, **kwargs):
    """
    Process(function, *args, **kwargs)

    Process the incoming MIDI event using a Python function, then continue
    executing the mididings patch with the events returned from that
    function.

    :param function:
        a function, or any other callable object, that will be called with
        a :class:`~.MidiEvent` object as its first argument.

        The function's return value can be:
          - a single :class:`~.MidiEvent` object.
          - a list of :class:`~.MidiEvent` objects.
          - ``None`` (or an empty list).

        Instead of ``return``\\ ing :class:`~.MidiEvent` objects, *function*
        may also be a generator that ``yield``\\ s :class:`~.MidiEvent`
        objects.

    :param \\*args:
        optional positional arguments that will be passed to *function*.

    :param \\*\\*kwargs:
        optional keyword arguments that will be passed to *function*.


    Any other MIDI processing will be stalled until *function* returns,
    so this should only be used with functions that don't block.
    Use :func:`Call()` for tasks that may take longer and/or don't require
    returning any MIDI events.
    """
    if _get_config('backend') == 'jack-rt' and not _get_config('silent'):
        print("WARNING: using Process() with the 'jack-rt' backend"
              " is probably a bad idea")
    return _CallBase(_call_partial(function, args, kwargs, True), False, False)


@_overload.mark(
    """
    Call(function, *args, **kwargs)
    Call(thread=..., **kwargs)

    Schedule a Python function for execution.
    The incoming event is discarded.

    :param function:
        a function, or any other callable object.
        If the function accepts arguments, its first argument will be a copy
        of the :class:`~.MidiEvent` that triggered the function call.

        The function's return value is ignored.

    :param thread:
        like *function*, but causes the function to be run in its own thread.

    :param \\*args:
        optional positional arguments that will be passed to *function*.

    :param \\*\\*kwargs:
        optional keyword arguments that will be passed to *function*.
    """
)
@_unitrepr.accept(_collections_abc.Callable, None, kwargs={ None: None })
def Call(function, *args, **kwargs):
    return _CallBase(_call_partial(function, args, kwargs), True, False)

@_overload.mark
@_unitrepr.accept(_collections_abc.Callable, kwargs={ None: None })
def Call(thread, **kwargs):
    return _CallThread(_call_partial(thread, (), kwargs))


@_unitrepr.accept((str, _collections_abc.Callable))
def System(command):
    """
    Run an arbitrary shell command.
    The incoming event is discarded.

    :param command:
        a string which will be passed verbatim to the shell.

        Alternatively it may also be a Python function that accepts a
        :class:`~.MidiEvent` argument and returns the command string to
        be executed.
    """
    return _System(command)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from _mididings import __version__ as internal_version
from mididings.setup import config, hook
from mididings.engine import run, process_file
from mididings.constants import *
from mididings.scene import *
from mididings.units import *


__version__ = internal_version


import mididings.misc as _misc
__all__ = _misc.prune_globals(globals())



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import inspect
import functools
import itertools
import termios
import fcntl
import struct
import sys

import decorator


def flatten(arg):
    """
    Flatten nested sequences into a single list.
    """
    if issequence(arg):
        return list(itertools.chain(*(flatten(i) for i in arg)))
    else:
        return [arg]


def issequence(seq, accept_string=False):
    """
    Return whether seq is of a sequence type. By default, strings are not
    considered sequences.
    """
    if not accept_string and isinstance(seq, str):
        return False

    try:
        iter(seq)
        return True
    except TypeError:
        return False


def issequenceof(seq, t):
    """
    Return whether seq is a sequence with elements of type t.
    """
    return issequence(seq) and all(isinstance(v, t) for v in seq)


def islambda(f):
    lam = lambda: None
    return isinstance(f, type(lam)) and f.__name__ == lam.__name__


_argspec_cache = {}

def getargspec(f):
    """
    Wrapper around inspect.getargspec() that returns sensible results for
    functools.partial objects.
    All results are cached since inspect.getargspec() is a little slow.
    """
    if f in _argspec_cache:
        return _argspec_cache[f]
    else:
        if isinstance(f, functools.partial):
            argspec = list(inspect.getfullargspec(f.func))
            argspec[0] = argspec[0][len(f.args):]
            r = tuple(argspec)
        else:
            r = inspect.getfullargspec(f)
        _argspec_cache[f] = r
        return r

"""
The inspect.formatargspec() function was dropped in Python 3.11 but we need
it when constructing signature changing decorators based on result of
inspect.getargspec() or inspect.getfullargspec(). The code here implements
inspect.formatargspec() based on Parameter and Signature from inspect module,
which were added in Python 3.6. Thanks to Cyril Jouve for the implementation.
"""
def formatargspec(args, varargs=None, varkw=None, defaults=None,
                  kwonlyargs=(), kwonlydefaults={}, annotations={}):
    if kwonlydefaults is None:
        kwonlydefaults = {}
    ndefaults = len(defaults) if defaults else 0
    parameters = [
        inspect.Parameter(
            arg,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            default=defaults[i] if i >= 0 else inspect.Parameter.empty,
            annotation=annotations.get(arg, inspect.Parameter.empty),
        ) for i, arg in enumerate(args, ndefaults - len(args))
    ]
    if varargs:
        parameters.append(inspect.Parameter(varargs, inspect.Parameter.VAR_POSITIONAL))
    parameters.extend(
        inspect.Parameter(
            kwonlyarg,
            inspect.Parameter.KEYWORD_ONLY,
            default=kwonlydefaults.get(kwonlyarg, inspect.Parameter.empty),
            annotation=annotations.get(kwonlyarg, inspect.Parameter.empty),
        ) for kwonlyarg in kwonlyargs
    )
    if varkw:
        parameters.append(inspect.Parameter(varkw, inspect.Parameter.VAR_KEYWORD))
    return_annotation = annotations.get('return', inspect.Signature.empty)
    return str(inspect.Signature(parameters, return_annotation=return_annotation))

class deprecated(object):
    def __init__(self, replacement=None):
        self.replacement = None

    def wrapper(self, f, *args, **kwargs):
        # XXX: avoid circular import
        from mididings.setup import get_config

        if (not (hasattr(f, '_already_used') and f._already_used) and
                not get_config('silent')):
            if self.replacement:
                print(f"{f.__name__}() is deprecated, please use {self.replacement}() instead")
            else:
                print(f"{f.__name__}() is deprecated")
            f._already_used = True
        return f(*args, **kwargs)

    def __call__(self, f):
        f._deprecated = True
        return decorator.decorator(self.wrapper, f)


class NamedFlag(int):
    """
    An integer type where each value has a name attached to it.
    """
    def __new__(cls, value, name):
        return int.__new__(cls, value)
    def __init__(self, value, name):
        self.name = name
    def __getnewargs__(self):
        return (int(self), self.name)
    def __repr__(self):
        return self.name
    def __str__(self):
        return self.name


class NamedBitMask(NamedFlag):
    """
    Like NamedFlag, but bit operations | and ~ are also reflected in the
    resulting value's string representation.
    """
    def __or__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return type(self)(
            int(self) | int(other),
            '%s|%s' % (self.name, other.name)
        )
    def __invert__(self):
        return type(self)(
            ~int(self) & ((1 << 30) - 1),
            ('~%s' if '|' not in self.name else '~(%s)') % self.name
        )


def prune_globals(g):
    return [n for (n, m) in g.items()
        if not inspect.ismodule(m)
        and not n.startswith('_')
        #and not (hasattr(m, '_deprecated'))
    ]


def sequence_to_hex(data):
    return ' '.join(hex(x)[2:].zfill(2) for x in data)


class bytestring(object):
    def __init__(self, data):
        self.data = data
    def __repr__(self):
        return '\'%s\'' % ''.join('\\x' + hex(x)[2:].zfill(2)
                                    for x in self.data)


def get_terminal_size():
    """
    Return the height and width of the terminal.
    """
    try:
        s = struct.pack("HHHH", 0, 0, 0, 0)
        fd = sys.stdout.fileno()
        x = fcntl.ioctl(fd, termios.TIOCGWINSZ, s)
        t = struct.unpack("HHHH", x)
        return t[0], t[1]
    except Exception:
        return 25, 80



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m
import mididings.engine as _engine
import mididings.event as _event
import mididings.util as _util


def _panic_bypass():
    # send all notes off (CC #123) and sustain off (CC #64) to all output
    # ports and on all channels
    for p in _engine.out_ports():
        for c in range(16):
            _engine.output_event(
                _event.CtrlEvent(p, _util.offset(c), 123, 0))
            _engine.output_event(
                _event.CtrlEvent(p, _util.offset(c), 64, 0))


def Panic(bypass=True):
    """
    Generate all-notes-off (CC #123) and sustain off (CC #64) events on
    all channels.

    :param bypass:
        If true, events will be sent directly on all output ports, instead
        of originating from the :func:`Panic()` unit and being subject to
        further processing.
    """
    if bypass:
        return _m.Call(lambda ev: _panic_bypass())
    else:
        return _m.Fork([
            (_m.Ctrl(p, _util.offset(c), 123, 0) //
             _m.Ctrl(p, _util.offset(c), 64, 0))
                for p in _engine.out_ports()
                for c in range(16)
        ])



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

import mididings.units as _units
import mididings.constants as _constants


class Patch(_mididings.Patch):
    def __init__(self, p):
        _mididings.Patch.__init__(self, self.build(p))

    def build(self, p):
        if isinstance(p, _units.base._Chain):
            return Patch.Chain(self.build(i) for i in p)

        elif isinstance(p, list):
            gen = (self.build(i) for i in p)

            remove_duplicates = True
            if hasattr(p, 'remove_duplicates'):
                remove_duplicates = (p.remove_duplicates != False)

            return Patch.Fork(gen, remove_duplicates)

        elif isinstance(p, dict):
            return self.build(
                _units.splits._make_split(_units.base.Filter, p, unpack=True)
            )

        elif isinstance(p, _units.init._InitExit):
            return Patch.Single(_mididings.Pass(False))

        elif isinstance(p, _units.base._Unit):
            if isinstance(p.unit, _mididings.Unit):
                return Patch.Single(p.unit)
            elif isinstance(p.unit, _mididings.UnitEx):
                return Patch.Extended(p.unit)

        elif isinstance(p, _constants._EventType):
            return Patch.Single(_mididings.TypeFilter(p))

        raise TypeError(f"type '{type(p).__name__}' not allowed in patch. offending object is: {p!r}")


def get_init_patches(patch):
    if isinstance(patch, _units.base._Chain):
        return flatten([get_init_patches(p) for p in patch])

    elif isinstance(patch, list):
        return flatten([get_init_patches(p) for p in patch])

    elif isinstance(patch, dict):
        return flatten([get_init_patches(p) for p in patch.values()])

    elif isinstance(patch, _units.init._InitExit):
        return [patch.init_patch]

    else:
        return []


def get_exit_patches(patch):
    if isinstance(patch, _units.base._Chain):
        return flatten([get_exit_patches(p) for p in patch])

    elif isinstance(patch, list):
        return flatten([get_exit_patches(p) for p in patch])

    elif isinstance(patch, dict):
        return flatten([get_exit_patches(p) for p in patch.values()])

    elif isinstance(patch, _units.init._InitExit):
        return [patch.exit_patch]

    else:
        return []


def flatten(patch):
    r = []
    for i in patch:
        if isinstance(i, list) and not isinstance(i, _units.base._Chain):
            r.extend(i)
        else:
            r.append(i)
    return r



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings.misc as _misc
import mididings.constants as _constants
import mididings.setup as _setup

import sys as _sys


_NOTE_NUMBERS = {
    'c':   0,
    'c#':  1, 'db':  1,
    'd':   2,
    'd#':  3, 'eb':  3,
    'e':   4,
    'f':   5,
    'f#':  6, 'gb':  6,
    'g':   7,
    'g#':  8, 'ab':  8,
    'a':   9,
    'a#': 10, 'bb': 10,
    'b':  11,
}

_NOTE_NAMES = {
     0: 'c',
     1: 'c#',
     2: 'd',
     3: 'd#',
     4: 'e',
     5: 'f',
     6: 'f#',
     7: 'g',
     8: 'g#',
     9: 'a',
    10: 'a#',
    11: 'b',
}

_CONTROLLER_NAMES = {
     0: 'Bank select (MSB)',
     1: 'Modulation',
     6: 'Data entry (MSB)',
     7: 'Volume',
    10: 'Pan',
    11: 'Expression',
    32: 'Bank select (LSB)',
    38: 'Data entry (LSB)',
    64: 'Sustain',
    65: 'Portamento',
    66: 'Sostenuto',
    67: 'Soft pedal',
    68: 'Legato pedal',
    98: 'NRPN (LSB)',
    99: 'NRPN (MSB)',
   100: 'RPN (LSB)',
   101: 'RPN (MSB)',
   121: 'Reset all controllers',
   123: 'All notes off',
}


def note_number(note, allow_end=False):
    """
    note_number(note)

    Convert note name to note number.

    :param note:
        any valid :ref:`note description <notes>` (name or number).
    :return:
        MIDI note number.
    """
    if isinstance(note, int):
        r = note
    elif isinstance(note, str):
        note = note.lower()
        # find first digit
        for i in range(len(note)):
            if note[i].isdigit() or note[i] == '-':
                break
        try:
            name = note[:i]
            octave = int(note[i:])
            r = (_NOTE_NUMBERS[name] +
                    (octave + _setup.get_config('octave_offset')) * 12)
        except Exception:
            raise ValueError(f"invalid note name '{note}'")
    else:
        raise TypeError("note must be an integer or string")

    end = 128 if not allow_end else 129
    if not (0 <= r < end):
        raise ValueError(f"note number {r} is out of range")
    return r


def note_limit(note):
    return note_number(note, allow_end=True)


def note_range(notes):
    """
    Convert note range to note numbers.

    :param notes:
        any valid :ref:`note range <notes>`
        (names or numbers, tuple or string).
        If this is a single note, the range containing only that note is
        returned.
    :return:
        tuple of two MIDI note numbers.
    """
    try:
        # single note
        n = note_number(notes)
        return (n, n + 1)
    except Exception:
        try:
            if isinstance(notes, tuple):
                # tuple of note numbers
                return note_limit(notes[0]), note_limit(notes[1])
            elif isinstance(notes, str):
                # note range string
                nn = notes.split(':', 1)
                lower = note_limit(nn[0]) if nn[0] else 0
                upper = note_limit(nn[1]) if nn[1] else 0
                return lower, upper
            else:
                raise TypeError("note range must be a tuple of integers or a string")
        except (ValueError, IndexError):
            raise ValueError(f"invalid note range {notes!r}")


def note_name(note):
    """
    Get note name from note number.

    :param note:
        a MIDI note number.
    :return:
        note name as a string.
    """
    if not isinstance(note, int):
        raise TypeError("note must be an integer")
    return _NOTE_NAMES[note % 12] + str(
                (note // 12) - _setup.get_config('octave_offset'))


def tonic_note_number(key):
    return _NOTE_NUMBERS[key]


def controller_name(ctrl):
    """
    Get controller description.
    """
    if ctrl in _CONTROLLER_NAMES:
        return _CONTROLLER_NAMES[ctrl]
    else:
        return None


def event_type(type):
    if type not in _constants._EVENT_TYPES:
        raise ValueError(f"invalid event type {type!r}")
    return type


def port_number(port):
    """
    Convert port description to port number.

    :param port:
        a :ref:`port name <ports>` or number.
    :return:
        the port's number.
    """
    if isinstance(port, int):
        if actual(port) < 0:
            raise ValueError(f"invalid port number {port}")
        return port
    elif isinstance(port, str):
        in_ports = _setup._in_portnames
        out_ports = _setup._out_portnames

        if (port in in_ports and port in out_ports and
                in_ports.index(port) != out_ports.index(port)):
            raise ValueError(f"port name '{port}' is ambiguous")
        elif port in in_ports:
            return offset(in_ports.index(port))
        elif port in out_ports:
            return offset(out_ports.index(port))
        else:
            raise ValueError(f"invalid port name '{port}'")
    else:
        raise TypeError("port must be an integer or string")


def channel_number(channel):
    if not isinstance(channel, int):
        raise TypeError("channel must be an integer")
    if not (0 <= actual(channel) < 16):
        raise ValueError(f"channel number {channel} is out of range")
    return channel


def program_number(program):
    if not isinstance(program, int):
        raise TypeError("program must be an integer")
    if not (0 <= actual(program) < 128):
        raise ValueError(f"program number {program} is out of range")
    return program


def ctrl_number(ctrl):
    if not isinstance(ctrl, int):
        raise TypeError("controller must be an integer")
    if not (0 <= ctrl < 128):
        raise ValueError(f"controller number {ctrl} is out of range")
    return ctrl


def ctrl_value(value, allow_end=False):
    if not isinstance(value, int):
        raise TypeError("controller value must be an integer")
    end = 128 if not allow_end else 129
    if not (0 <= value < end):
        raise ValueError(f"controller value {value} is out of range")
    return value

def ctrl_limit(value):
    return ctrl_value(value, allow_end=True)

def ctrl_range(value):
    try:
        n = ctrl_value(value)
        return (n, n + 1)
    except Exception:
        if isinstance(value, tuple) and len(value) == 2:
            return (ctrl_limit(value[0]), ctrl_limit(value[1]))
    raise ValueError(f"invalid controller value range {value!r}")


def velocity_value(velocity, allow_end=False):
    if not isinstance(velocity, int):
        raise TypeError("velocity must be an integer")
    end = 128 if not allow_end else 129
    if not (0 <= velocity < end):
        raise ValueError(f"velocity {velocity} is out of range")
    return velocity

def velocity_limit(velocity):
    return velocity_value(velocity, allow_end=True)

def velocity_range(velocity):
    try:
        n = velocity_value(velocity)
        return (n, n + 1)
    except Exception:
        if isinstance(velocity, tuple) and len(velocity) == 2:
            return (velocity_limit(velocity[0]), velocity_limit(velocity[1]))
    raise ValueError(f"invalid velocity range {velocity!r}")


def scene_number(scene):
    if not isinstance(scene, int):
        raise TypeError("scene number must be an integer")
    if actual(scene) < 0:
        raise ValueError(f"scene number {scene} is out of range")
    return scene


def subscene_number(subscene):
    if not isinstance(subscene, int):
        raise TypeError("subscene number must be an integer")
    if actual(subscene) < 0:
        raise ValueError(f"subscene number {subscene} is out of range")
    return subscene


def sysex_to_bytearray(sysex):
    if isinstance(sysex, str):
        if sysex.startswith('F0') or sysex.startswith('f0'):
            if len(sysex) < 3 or sysex[2] not in ', \f\n\r\t\v':
                return bytearray(int(sysex[i:i+2], 16)
                                 for i in range(0, len(sysex), 2))
            else:
                return bytearray(int(x, 16) for x in sysex.split(sysex[2]))
        else:
            return bytearray(map(ord, sysex))
    elif isinstance(sysex, bytearray):
        return sysex
    else:
        return bytearray(sysex)


def sysex_data(sysex, allow_partial=False):
    sysex = sysex_to_bytearray(sysex)
    if len(sysex) < 2:
        raise ValueError("sysex too short")
    elif sysex[0] != 0xf0:
        raise ValueError("sysex doesn't start with F0")
    elif sysex[-1] != 0xf7 and not allow_partial:
        raise ValueError("sysex doesn't end with F7")
    else:
        for c in sysex[1:-1]:
            if c > 0x7f:
                raise ValueError(f"sysex data byte {c:x} is out of range")
    return sysex


def sysex_manufacturer(manufacturer):
    if not _misc.issequence(manufacturer, True):
        manufacturer = [manufacturer]
    manid = sysex_to_bytearray(manufacturer)
    if len(manid) not in (1, 3):
        raise ValueError("manufacturer id must be either one or three bytes")
    elif len(manid) == 3 and manid[0] != 0x00:
        raise ValueError("three-byte manufacturer id must start with 0x00")
    else:
        for c in manid:
            if c > 0x7f:
                raise ValueError(f"manufacturer id byte {c:x} is out of range")
    return manid



class NoDataOffset(int):
    """
    An integer type that's unaffected by data offset conversions.
    """
    def __new__(cls, value):
        return int.__new__(cls, value)
    def __repr__(self):
        return f"NoDataOffset({int(self)})"
    def __str__(self):
        return f"NoDataOffset({int(self)})"


def offset(n):
    """
    Add current data offset.
    """
    return n + _setup.get_config('data_offset')


def actual(n):
    """
    Subtract current data offset to get the "real" value used on the C++ side.
    """
    if isinstance(n, NoDataOffset):
        return int(n)
    else:
        return n - _setup.get_config('data_offset')



# define wrappers around parameter check functions that will also accept
# event attribute references
def _allow_event_attribute(f):
    def func(first, *args, **kwargs):
        if isinstance(first, _constants._EventAttribute):
            return first
        else:
            return f(first, *args, **kwargs)
    return func

port_number_ref     = _allow_event_attribute(port_number)
channel_number_ref  = _allow_event_attribute(channel_number)
note_number_ref     = _allow_event_attribute(note_number)
velocity_value_ref  = _allow_event_attribute(velocity_value)
ctrl_number_ref     = _allow_event_attribute(ctrl_number)
ctrl_value_ref      = _allow_event_attribute(ctrl_value)
program_number_ref  = _allow_event_attribute(program_number)
scene_number_ref    = _allow_event_attribute(scene_number)
subscene_number_ref = _allow_event_attribute(subscene_number)
actual_ref          = _allow_event_attribute(actual)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings.units.call import _CallBase

import mididings.overload as _overload
import mididings.unitrepr as _unitrepr
import mididings.constants as _constants
import mididings.misc as _misc

import sys as _sys
import collections as _collections
import collections.abc as _collections_abc


class _Print(_CallBase):
    max_name_length = -1
    max_portname_length = -1
    portnames_used = False

    def __init__(self, name, portnames):
        if portnames is not None:
            _Print.portnames_used = True

        # find maximum name length
        if name:
            _Print.max_name_length = max(_Print.max_name_length, len(name))

        # using a separare object to do the actual printing avoids keeping
        # references back to this unit, thus preventing a cycle
        printer = _Printer(name, portnames)

        _CallBase.__init__(self, printer, True, True)


class _Printer(object):
    def __init__(self, name, portnames):
        self.name = name
        self.portnames = portnames
        # to be calculated later
        self.ports = None

    def __call__(self, ev):
        # lazy import to avoid problems with circular imports
        from mididings import engine

        # get list of port names to be used (delayed 'til first use,
        # because the engine doesn't yet exist during __init__)
        if self.ports is None:
            if self.portnames == 'in':
                self.ports = engine.in_ports()
            elif self.portnames == 'out':
                self.ports = engine.out_ports()
            else:
                self.ports = []

        # find maximum port name length (delayed for the same reason as above)
        if _Print.portnames_used and _Print.max_portname_length == -1:
            all_ports = engine.in_ports() + engine.out_ports()
            _Print.max_portname_length = max(len(p) for p in all_ports)

        if self.name:
            namestr = '%-*s ' % (_Print.max_name_length + 1, self.name + ':')
        elif _Print.max_name_length != -1:
            # no name, but names used elsewhere, so indent appropriately
            namestr = ' ' * (_Print.max_name_length + 2)
        else:
            namestr = ''

        if ev.type == _constants.SYSEX:
            eventmax = _misc.get_terminal_size()[1] - len(namestr)
        else:
            eventmax = 0
        eventstr = ev.to_string(self.ports, _Print.max_portname_length,
                                eventmax)

        print(f"{namestr}{eventstr}")


class _PrintString(_CallBase):
    def __init__(self, string):
        self.string = string
        _CallBase.__init__(self, self.do_print, True, True)
    def do_print(self, ev):
        string = (self.string(ev) if hasattr(self.string, '__call__')
                                else self.string)
        print(string)


@_overload.mark(
    """
    Print(name=None, portnames=None)
    Print(string=...)

    Print event data or strings to the console.

    :param name:
        an optional identifier that will be printed before any event data,
        in order to distinguish between multiple :func:`Print()` units
        in the same patch.

    :param portnames:
        determines which port names will be printed:

        - ``'in'``: the input port name corresponding to the event's
          port number.
        - ``'out'``: the output port name corresponding to the event's
          port number.
        - ``None``: port number only.

    :param string:
        print a fixed string.
        Alternatively *string* may also be a Python function that accepts
        a :class:`~.MidiEvent` argument and returns the string to be printed.
    """
)
@_unitrepr.accept((str, type(None)), ('in', 'out', None))
def Print(name=None, portnames=None):
    return _Print(name, portnames)

@_overload.mark
@_unitrepr.accept((str, _collections_abc.Callable))
def Print(string):
    return _PrintString(string)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

from mididings.misc import NamedFlag as _NamedFlag
from mididings.misc import NamedBitMask as _NamedBitMask


class _EventType(_NamedBitMask):
    pass


class _EventAttribute(_NamedFlag):
    pass


_EVENT_TYPES = {}

# populate this module with midi event type constants
for _name, _value in _mididings.MidiEventType.names.items():
    _type_object = _EventType(int(_value), _name)

    # add event type object to this module's global namespace
    globals()[_name] = _type_object

    # only masks matching a single event type (exactly one bit set) are added
    # to the event types dict
    if len([_x for _x in range(32) if _value & (1 << _x)]) == 1:
        _EVENT_TYPES[int(_value)] = _type_object


# populate this module with midi event attribute constants
for _name, _value in _mididings.EventAttribute.names.items():
    _attribute_object = _EventAttribute(int(_value), "EVENT_" + _name)

    # add event attribute object to this module's global namespace
    globals()["EVENT_" + _name] = _attribute_object



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings import misc

import inspect
import collections
import collections.abc
import types
import functools

import decorator


class accept(object):
    """
    A decorator that applies type checks and other constraints to the
    arguments of the decorated function.

    Constraints must be given in the order of the function's positional
    arguments, one constraint per argument. If the function accepts
    variable arguments, one additional constraint must be specified, and
    will be applied to each of those arguments.

    If the optional keyword argument add_varargs is True, the decorated
    function will accept variable arguments, which are combined with the
    last regular positional argument into a single tuple. This tuple is
    then passed to the original function as a single argument.

    The kwargs dictionary, itself passed as an optional keyword argument,
    maps keyword arguments to the corresponding constraints. None can be
    used as a key in this dictionary to specify constraints for keyword
    arguments with any name.
    """
    def __init__(self, *constraints, **kwargs):
        self.add_varargs = (kwargs['add_varargs']
                            if 'add_varargs' in kwargs else False)
        kwargs_constraints = kwargs['kwargs'] if 'kwargs' in kwargs else {}

        # build all constraints
        self.constraints = [_make_constraint(c) for c in constraints]
        self.kwargs_constraints = dict((k, _make_constraint(v))
                                        for k, v in kwargs_constraints.items())

    def __call__(self, f):
        argspec = misc.getargspec(f)
        self.arg_names = argspec[0]
        self.have_varargs = (argspec[1] is not None) and not self.add_varargs

        assert ((len(self.constraints) == len(self.arg_names)
                    and not self.have_varargs)
             or (len(self.constraints) == len(self.arg_names) + 1
                    and self.have_varargs))

        if self.add_varargs:
            # add varargs to the signature and use decorator.FunctionMaker
            # directly to create the decorated function
            orig_signature = inspect.getfullargspec(f)
            assert orig_signature.varargs is None
            signature = orig_signature._replace(varargs='args')
            evaldict = self.wrapper.__globals__.copy()
            evaldict['_call_'] = self.wrapper
            evaldict['_func_'] = f
            return decorator.FunctionMaker.create(
                f"{f.__name__}{misc.formatargspec(*signature)}",
                'return _call_(_func_, %(shortsignature)s)',
                evaldict, undecorated=f, __wrapped__=f, doc=f.__doc__)
        else:
            return decorator.decorator(self.wrapper, f)

    def wrapper(self, f, *args, **kwargs):
        mod_args = []
        mod_kwargs = {}

        for constraint, arg_name, arg in zip(self.constraints,
                                             self.arg_names, args):
            if self.add_varargs and arg_name == self.arg_names[-1]:
                # add_varargs is True and this is the last argument:
                # combine with varargs
                index = len(self.arg_names)
                if isinstance(arg, types.GeneratorType):
                    arg = tuple(arg) + args[index:]
                else:
                    arg = (arg,) + args[index:]

            a = _apply_constraint(constraint, arg, f.__name__, arg_name)
            mod_args.append(a)

        if self.have_varargs:
            index = len(self.arg_names)
            constraint = self.constraints[index]

            for arg in args[index:]:
                a = _apply_constraint(constraint, arg, f.__name__, None)
                mod_args.append(a)

        for k, v in kwargs.items():
            if k in self.kwargs_constraints:
                a = _apply_constraint(self.kwargs_constraints[k],
                                      v, f.__name__, k)
            elif None in self.kwargs_constraints:
                a = _apply_constraint(self.kwargs_constraints[None],
                                      v, f.__name__, k)
            else:
                raise TypeError(f"{f.__name__}() got an unexpected keyword argument '{k}'")
            mod_kwargs[k] = a

        return f(*mod_args, **mod_kwargs)


def _apply_constraint(constraint, arg, func_name, arg_name):
    try:
        return constraint(arg)
    except (TypeError, ValueError) as ex:
        typestr = "type" if isinstance(ex, TypeError) else "value"
        argstr = ((f"for parameter '{arg_name}'")
                        if arg_name else "in varargs")

        raise type(ex)(f"invalid {typestr} {argstr} of function {func_name}():\n{str(ex)}")


def _make_constraint(c):
    if c is None:
        return _any()
    elif isinstance(c, list):
        if len(c) == 1:
            # single-item list: sequenceof() constraint
            return sequenceof(c[0])
        else:
            # list: tupleof() constraint
            return tupleof(*c)
    elif isinstance(c, dict):
        assert len(c) == 1
        # single-item dict: mappingof() constraint
        return mappingof(list(c.keys())[0], list(c.values())[0])
    elif isinstance(c, tuple):
        if all(inspect.isclass(cc) for cc in c):
            # multiple types: type constraint
            return _type_constraint(c, True)
        else:
            # any other tuple: value constraint
            return _value_constraint(c)
    elif inspect.isclass(c):
        # single type: type constraint
        return _type_constraint(c)
    elif isinstance(c, _constraint):
        # constraint object
        return c
    elif isinstance(c, collections.abc.Callable):
        # function or other callable object
        return transform(c)
    else:
        assert False



class _constraint(object):
    pass


class _any(_constraint):
    def __call__(self, arg):
        return arg


class _type_constraint(_constraint):
    def __init__(self, types, multiple=False):
        self.types = types
        self.multiple = multiple

    def __call__(self, arg):
        if not self.multiple:
            # single type, check if instance
            if not isinstance(arg, self.types):
                raise TypeError(f"expected {self.types.__name__}, got {type(arg).__name__}")
            return arg
        else:
            # multiple types, check if instance
            if not isinstance(arg, self.types):
                argtypes = ", ".join(c.__name__ for c in self.types)
                raise TypeError(f"expected one of ({argtypes}), got {type(arg).__name__}")
            return arg

    def __repr__(self):
        if not self.multiple:
            return self.types.__name__
        else:
            return repr(tuple(map(_type_constraint, self.types)))


class _value_constraint(_constraint):
    def __init__(self, values):
        self.values = values

    def __call__(self, arg):
        if arg not in self.values:
            args = ", ".join(repr(c) for c in self.values)
            raise ValueError(f"expected one of ({args}), got {arg!r}")
        return arg

    def __repr__(self):
        return repr(tuple(self.values))


class nullable(_constraint):
    """
    Allows the argument to be None, instead of matching any other constraint.
    """
    def __init__(self, what):
        self.what = _make_constraint(what)

    def __call__(self, arg):
        if arg is None:
            return None
        return self.what(arg)

    def __repr__(self):
        return f"nullable({self.what!r})"


class sequenceof(_constraint):
    """
    Checks that the argument is a sequence, and applies the same constraint to
    each element in that sequence.
    """
    def __init__(self, what):
        self.what = _make_constraint(what)

    def __call__(self, arg):
        if not misc.issequence(arg):
            raise TypeError("not a sequence")
        try:
            t = type(arg) if not isinstance(arg, types.GeneratorType) else list
            return t(self.what(value) for value in arg)
        except (TypeError, ValueError) as ex:
            raise type(ex)(f"illegal item in sequence: {str(ex)}")

    def __repr__(self):
        return repr([self.what])


class tupleof(_constraint):
    """
    Checks that the argument is a sequence of a fixed length, and applies
    different constraints to each element in that sequence.
    """
    def __init__(self, *what):
        self.what = [_make_constraint(c) for c in what]

    def __call__(self, arg):
        if not misc.issequence(arg):
            raise TypeError("not a sequence")
        if len(arg) != len(self.what):
            raise ValueError(f"expected sequence of {len(self.what)} items, got {len(arg)}")
        try:
            t = type(arg) if not isinstance(arg, types.GeneratorType) else list
            return t(what(value) for what, value in zip(self.what, arg))
        except (TypeError, ValueError) as ex:
            raise type(ex)(f"illegal item in sequence: {str(ex)}")

    def __repr__(self):
        return repr(self.what)


class mappingof(_constraint):
    """
    Checks that the argument is a dictionary, and applies constraints to each
    key and each value.
    """
    def __init__(self, fromwhat, towhat):
        self.fromwhat = _make_constraint(fromwhat)
        self.towhat = _make_constraint(towhat)

    def __call__(self, arg):
        if not isinstance(arg, dict):
            raise TypeError("not a dictionary")
        try:
            keys = (self.fromwhat(key) for key in arg.keys())
        except (TypeError, ValueError) as ex:
            raise type(ex)(f"illegal key in dictionary: {str(ex)}")
        try:
            values = (self.towhat(value) for value in arg.values())
        except (TypeError, ValueError) as ex:
            raise type(ex)(f"illegal value in dictionary: {str(ex)}")
        return dict(zip(keys, values))

    def __repr__(self):
        return repr({self.fromwhat: self.towhat})


class flatten(_constraint):
    """
    Flattens all arguments into a single list, and applies a constraint to
    each element in that list.
    """
    def __init__(self, what, return_type=None):
        self.what = _make_constraint(what)
        self.return_type = return_type

    def __call__(self, arg):
        try:
            r = [self.what(value) for value in misc.flatten(arg)]
            return r if self.return_type is None else self.return_type(r)
        except (TypeError, ValueError) as ex:
            raise type(ex)(f"illegal item in sequence: {str(ex)}")

    def __repr__(self):
        return f"flatten({self.what!r})"


class each(_constraint):
    """
    Applies each of the given constraints.
    """
    def __init__(self, *requirements):
        self.requirements = [_make_constraint(c) for c in requirements]

    def __call__(self, arg):
        for what in self.requirements:
            arg = what(arg)
        return arg

    def __repr__(self):
        return f"each({', '.join(repr(c) for c in self.requirements)})"


class either(_constraint):
    """
    Accepts the argument if any of the given constraints can be applied.
    """
    def __init__(self, *alternatives):
        self.alternatives = [_make_constraint(c) for c in alternatives]

    def __call__(self, arg):
        errors = []
        for n, what in enumerate(self.alternatives):
            try:
                return what(arg)
            except (TypeError, ValueError) as ex:
                exstr = str(ex).replace('\n', '\n    ')
                errors.append(f"    #{n + 1} {what}: {type(ex).__name__}: {exstr}")
        raise TypeError("none of the alternatives matched:\n" +
                        '\n'.join(errors))

    def __repr__(self):
        return f"either({', '.join(repr(c) for c in self.alternatives)})"


class transform(_constraint):
    """
    Applies a function to its argument.
    """
    def __init__(self, function):
        self.function = function

    def __call__(self, arg):
        return self.function(arg)

    def __repr__(self):
        return _function_repr(self.function)


class condition(_constraint):
    """
    Accepts the argument if the given function returns True.
    """
    def __init__(self, function):
        self.function = function

    def __call__(self, arg):
        if not self.function(arg):
            raise ValueError(f"condition not met: {_function_repr(self.function)}")
        return arg

    def __repr__(self):
        return f"condition({_function_repr(self.function)})"


class reduce_bitmask(_constraint):
    """
    Flattens all arguments and reduces them to a single bitmask.
    """
    def __init__(self, what):
        self.what = _make_constraint(what)

    def __call__(self, arg):
        seq = self.what(misc.flatten(arg))
        return functools.reduce(lambda x, y: x | y, seq)


def _function_repr(f):
    if misc.islambda(f):
        s = inspect.getsource(f).strip()

        # inspect.getsource() returns the entire line of code.
        # try to extract only the actual definition of the lambda function.
        # this will fail if more than one lambda function is defined on
        # the same line (and possibly for several other reasons...)
        start = s.index('lambda')
        end = len(s)
        parens = 0
        for n, c in enumerate(s[start:]):
            if parens == 0 and c in ',)]}':
                end = start + n
                break
            elif c in '([{':
                parens += 1
            elif c in ')]}':
                parens -= 1
        return s[start:end]
    else:
        return f.__name__



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m
import mididings.event as _event
import mididings.engine as _engine
from mididings.extra.per_channel import PerChannel as _PerChannel


class _VoiceFilter(object):
    def __init__(self, voice, time, retrigger):
        self.voice = voice
        self.time = time
        self.retrigger = retrigger

        # all notes currently being played
        self.notes = {}
        # note number this voice is playing
        self.current_note = None
        # if we had to fall back to a different voice
        self.diverted = False

    def __call__(self, ev):
        t = _engine.time()

        if ev.type == _m.NOTEON:
            # store new note, its velocity, and its time
            self.notes[ev.note] = (ev.velocity, t)
        elif ev.type == _m.NOTEOFF:
            # delete released note
            try:
                del self.notes[ev.note]
            except KeyError:
                # ignore unmatched note-offs (just in case...)
                pass

        sorted_notes = sorted(self.notes.keys())

        if len(sorted_notes):
            try:
                n = sorted_notes[self.voice]
                d = False
            except IndexError:
                # use the next best note:
                # lowest for negative index, otherwise highest
                n = sorted_notes[0 if self.voice < 0 else -1]
                d = True
        else:
            n = None
            d = False

        dt = ((t - self.notes[self.current_note][1])
                if self.current_note in self.notes else 0.0)

        # change current note if...
        if (
            # note number changed and...
            n != self.current_note and (
                # we're always retriggering notes
                self.retrigger or
                # lowest/heighest voice are a bit of a special case
                self.voice in (0, -1) or
                # current note is no longer held
                self.current_note not in self.notes or
                # our previous note is very recent
                (ev.type == _m.NOTEON and dt < self.time) or
                # the new note is "better" than previous one
                self.diverted and not d
        )):
            # yield note-off for previous note (if any)
            if self.current_note:
                yield _event.NoteOffEvent(
                            ev.port, ev.channel, self.current_note, 0)
                self.current_note = None

            dt = (t - self.notes[n][1]) if n in self.notes else 0.0

            # yield note-on for new note (if any)
            if n is not None and (
                # this is the note being played right now
                ev.note == n or
                # we're retriggering whenever a key is pressed or released
                self.retrigger or
                # our previous note is very recent
                dt < self.time
            ):
                yield _event.NoteOnEvent(
                            ev.port, ev.channel, n, self.notes[n][0])
                self.current_note = n
                self.diverted = d


def VoiceFilter(voice='highest', time=0.1, retrigger=False):
    """
     Filter individual voices from a chord.

    :param voice:
        The voice to be filtered.

        This can be ``'highest'``, ``'lowest'``, or an integer index
        (positive or negative, the same way Python lists are indexed,
        with 0 being the lowest voice and -1 the highest).

    :param time:
        The period in seconds for which a newly played note may still be
        unassigned from the selected voice, when additional notes are played.

    :param retrigger:
        If true, a new note-on event will be sent when a note is reassigned
        to the selected voice as a result of another note being released.
    """
    if voice == 'highest':
        voice = -1
    elif voice == 'lowest':
        voice = 0

    return _m.Filter(_m.NOTE) % _m.Process(_PerChannel(
        lambda: _VoiceFilter(voice, time, retrigger))
    )


def VoiceSplit(patches, fallback='highest', time=0.1, retrigger=False):
    """
    Create multiple :func:`VoiceFilter()` units to route each voice to
    a different instrument.

    :param patches:
        A list of patches, one for each voice.

    :param fallback:
        Which note to double when the number of notes played is less than
        the number of voices. Can be ``'highest'`` or ``'lowest'``.

    Regardless of the number of voices specified, the lowest and highest
    note played will always be routed to the first and last patch in
    the list, respectively.
    """
    vf = lambda n: VoiceFilter(n, time, retrigger)

    if fallback == 'lowest':
        return _m.Fork(
            [ vf( 0) >> patches[ 0] ] +
            [ vf( n) >> patches[ n] for n in range(-len(patches) + 1, 0) ]
        )
    else: # highest
        return _m.Fork(
            [ vf( n) >> patches[ n] for n in range(len(patches) - 1) ] +
            [ vf(-1) >> patches[-1] ]
        )



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings.util as _util

(
    # 1
    ACOUSTIC_GRAND_PIANO,
    BRIGHT_ACOUSTIC_PIANO,
    ELECTRIC_GRAND_PIANO,
    HONKY_TONK_PIANO,
    ELECTRIC_PIANO_1,
    ELECTRIC_PIANO_2,
    HARPSICHORD,
    CLAVINET,
    # 9
    CELESTA,
    GLOCKENSPIEL,
    MUSIC_BOX,
    VIBRAPHONE,
    MARIMBA,
    XYLOPHONE,
    TUBULAR_BELLS,
    DULCIMER,
    # 17
    DRAWBAR_ORGAN,
    PERCUSSIVE_ORGAN,
    ROCK_ORGAN,
    CHURCH_ORGAN,
    REED_ORGAN,
    ACCORDION,
    HARMONICA,
    TANGO_ACCORDION,
    # 25
    ACOUSTIC_GUITAR_NYLON,
    ACOUSTIC_GUITAR_STEEL,
    ELECTRIC_GUITAR_JAZZ,
    ELECTRIC_GUITAR_CLEAN,
    ELECTRIC_GUITAR_MUTED,
    OVERDRIVEN_GUITAR,
    DISTORTION_GUITAR,
    GUITAR_HARMONICS,
    # 33
    ACOUSTIC_BASS,
    ELECTRIC_BASS_FINGER,
    ELECTRIC_BASS_PICK,
    FRETLESS_BASS,
    SLAP_BASS_1,
    SLAP_BASS_2,
    SYNTH_BASS_1,
    SYNTH_BASS_2,
    # 41
    VIOLIN,
    VIOLA,
    CELLO,
    CONTRABASS,
    TREMOLO_STRINGS,
    PIZZICATO_STRINGS,
    ORCHESTRAL_HARP,
    TIMPANI,
    # 49
    STRING_ENSEMBLE_1,
    STRING_ENSEMBLE_2,
    SYNTH_STRINGS_1,
    SYNTH_STRINGS_2,
    CHOIR_AAHS,
    VOICE_OOHS,
    SYNTH_VOICE,
    ORCHESTRA_HIT,
    # 57
    TRUMPET,
    TROMBONE,
    TUBA,
    MUTED_TRUMPET,
    FRENCH_HORN,
    BRASS_SECTION,
    SYNTHBRASS_1,
    SYNTHBRASS_2,
    # 65
    SOPRANO_SAX,
    ALTO_SAX,
    TENOR_SAX,
    BARITONE_SAX,
    OBOE,
    ENGLISH_HORN,
    BASSOON,
    CLARINET,
    # 73
    PICCOLO,
    FLUTE,
    RECORDER,
    PAN_FLUTE,
    BLOWN_BOTTLE,
    SHAKUHACHI,
    WHISTLE,
    OCARINA,
    # 81
    LEAD_1_SQUARE,
    LEAD_2_SAWTOOTH,
    LEAD_3_CALLIOPE,
    LEAD_4_CHIFF,
    LEAD_5_CHARANG,
    LEAD_6_VOICE,
    LEAD_7_FIFTHS,
    LEAD_8_BASS_LEAD,
    # 89
    PAD_1_NEW_AGE,
    PAD_2_WARM,
    PAD_3_POLYSYNTH,
    PAD_4_CHOIR,
    PAD_5_BOWED,
    PAD_6_METALLIC,
    PAD_7_HALO,
    PAD_8_SWEEP,
    # 97
    FX_1_RAIN,
    FX_2_SOUNDTRACK,
    FX_3_CRYSTAL,
    FX_4_ATMOSPHERE,
    FX_5_BRIGHTNESS,
    FX_6_GOBLINS,
    FX_7_ECHOES,
    FX_8_SCI_FI,
    # 105
    SITAR,
    BANJO,
    SHAMISEN,
    KOTO,
    KALIMBA,
    BAGPIPE,
    FIDDLE,
    SHANAI,
    # 113
    TINKLE_BELL,
    AGOGO,
    STEEL_DRUMS,
    WOODBLOCK,
    TAIKO_DRUM,
    MELODIC_TOM,
    SYNTH_DRUM,
    REVERSE_CYMBAL,
    # 121
    GUITAR_FRET_NOISE,
    BREATH_NOISE,
    SEASHORE,
    BIRD_TWEET,
    TELEPHONE_RING,
    HELICOPTER,
    APPLAUSE,
    GUNSHOT
) = map(_util.NoDataOffset, range(128))


CTRL_BANK_SELECT_MSB        = 0
CTRL_MODULATION             = 1
CTRL_DATA_ENTRY_MSB         = 6
CTRL_VOLUME                 = 7
CTRL_PAN                    = 10
CTRL_EXPRESSION             = 11
CTRL_BANK_SELECT_LSB        = 32
CTRL_DATA_ENTRY_LSB         = 38
CTRL_SUSTAIN                = 64
CTRL_PORTAMENTO             = 65
CTRL_SOSTENUTO              = 66
CTRL_SOFT_PEDAL             = 67
CTRL_LEGATO_PEDAL           = 68
CTRL_NRPN_LSB               = 98
CTRL_NRPN_MSB               = 99
CTRL_RPN_LSB                = 100
CTRL_RPN_MSB                = 101
CTRL_RESET_ALL_CONTROLLERS  = 121
CTRL_ALL_NOTES_OFF          = 123



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

from mididings.units.base import _Unit

import mididings.constants as _constants
import mididings.util as _util
import mididings.overload as _overload
import mididings.unitrepr as _unitrepr


@_unitrepr.store
def Sanitize():
    """
    Sanitize()

    Make sure the event is a valid MIDI message.
    Events with invalid port (output), channel, controller, program or note
    number are discarded. Note velocity and controller values are confined to
    the range 0-127.
    """
    return _Unit(_mididings.Sanitize())


@_overload.mark(
    """
    SceneSwitch(number=EVENT_PROGRAM)
    SceneSwitch(offset=...)

    Switch to another scene.

    *number* can be a fixed scene number, or one of the
    :ref:`Event Attribute <event-attributes>` constants to use a value from
    the incoming event's data.
    With no arguments, the program number of the incoming event (which should
    be a program change) will be used.

    *offset* can be a positive or negative value that will be added to the
    current scene number, allowing you to go forward or backward in the list
    of scenes.
    """
)
@_unitrepr.accept(_util.scene_number_ref)
def SceneSwitch(number=_constants.EVENT_PROGRAM):
    return _Unit(_mididings.SceneSwitch(_util.actual_ref(number), 0))

@_overload.mark
@_unitrepr.accept(int)
def SceneSwitch(offset):
    return _Unit(_mididings.SceneSwitch(0, offset))


@_overload.mark(
    """
    SubSceneSwitch(number=EVENT_PROGRAM)
    SubSceneSwitch(offset=..., wrap=True)

    Switches between subscenes within a scene group.

    *number* can be a fixed subscene number, or one of the
    :ref:`Event Attribute <event-attributes>` constants to use a value from
    the incoming event's data.
    With no arguments, the program number of the incoming event (which should
    be a program change) will be used.

    *offset* can be a positive or negative value that will be added to the
    current subscene number, allowing you to go forward or backward in the
    list of subscenes. If *wrap* is ``True`` you can loop through subscenes.
    """
)
@_unitrepr.accept(_util.subscene_number_ref)
def SubSceneSwitch(number=_constants.EVENT_PROGRAM):
    return _Unit(_mididings.SubSceneSwitch(_util.actual_ref(number), 0, False))

@_overload.mark
@_unitrepr.accept(int, bool)
def SubSceneSwitch(offset, wrap=True):
    return _Unit(_mididings.SubSceneSwitch(0, offset, wrap))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings.units.base import _Unit, Fork, Chain, _UNIT_TYPES
from mididings.units.generators import Program, Ctrl
from mididings.units.modifiers import Port, Channel

import mididings.unitrepr as _unitrepr
import mididings.util as _util

import functools as _functools
import copy as _copy


class _InitExit(_Unit):
    def __init__(self, init_patch=[], exit_patch=[]):
        self.init_patch = init_patch
        self.exit_patch = exit_patch


@_unitrepr.accept(_UNIT_TYPES)
def Init(patch):
    """
    Init(patch)

    Add *patch* to the init patch of the scene containing this unit, so that
    it will be executed when switching to this scene.
    The patch is added in parallel to whatever is already in the init patch.

    This unit does no event processing in the patch it is inserted in, and
    discards all events.
    """
    return _InitExit(init_patch=patch)


@_unitrepr.accept(_UNIT_TYPES)
def Exit(patch):
    """
    Exit(patch)

    Add *patch* to the exit patch of the scene containing this unit, so that
    it will be executed when leaving this scene, switching to a different one.
    The patch is added in parallel to whatever is already in the exit patch.

    This unit does no event processing in the patch it is inserted in, and
    discards all events.
    """
    return _InitExit(exit_patch=patch)


def Output(port=None, channel=None, program=None,
           volume=None, pan=None, expression=None, ctrls={}):
    """
    Output(port=None, channel=None, program=None, volume=None, pan=None, expression=None, ctrls={})

    Route incoming events to the specified *port* and *channel*.
    When switching to the scene containing this unit, a program change and/or
    arbitrary control changes can be sent.

    To send a bank select (CC #0 and CC #32) before the program change,
    *program* can be a tuple with two elements, where the first element is the
    bank number, and the second is the program number.

    Values for the commonly used controllers *volume* (#7), *pan* (#10) and
    *expression* (#11) can be specified directly. For all other controllers,
    the *ctrls* argument can contain a mapping from controller numbers to
    their respective values.

    If *port* or *channel* are ``None``, events will be routed to the first
    port/channel.
    """
    if port is None:
        port = _util.offset(0)
    if channel is None:
        channel = _util.offset(0)

    if isinstance(program, tuple):
        bank, program = program
    else:
        bank = None

    init = []

    if bank is not None:
        init.append(Ctrl(port, channel, 0, bank // 128))
        init.append(Ctrl(port, channel, 32, bank % 128))

    if program is not None:
        init.append(Program(port, channel, program))

    if volume is not None:
        init.append(Ctrl(port, channel, 7, volume))
    if pan is not None:
        init.append(Ctrl(port, channel, 10, pan))
    if expression is not None:
        init.append(Ctrl(port, channel, 11, expression))

    for k, v in ctrls.items():
        init.append(Ctrl(port, channel, k, v))

    return Fork([
        Init(init),
        Port(port) >> Channel(channel)
    ])


class OutputTemplate(object):
    """
    OutputTemplate(*args, **kwargs)

    Create an object that when called will behave like :func:`~.Output()`,
    with *args* and *kwargs* replacing some of its arguments.
    That is, :class:`~.OutputTemplate()` is not a unit by itself, but returns
    one when called.

    This works just like ``functools.partial(Output, *args, **kwargs)``, but
    with the added benefit that an :class:`~.OutputTemplate()` object also
    supports operator ``>>`` like any unit.
    """
    def __init__(self, *args, **kwargs):
        self.partial = _functools.partial(Output, *args, **kwargs)
        self.before = []
        self.after = []

    def __call__(self, *args, **kwargs):
        return (Chain(self.before)
                >> self.partial(*args, **kwargs)
                >> Chain(self.after))

    def __rshift__(self, other):
        if not isinstance(other, _UNIT_TYPES):
            return NotImplemented
        r = _copy.copy(self)
        r.after = self.after + [other]
        return r

    def __rrshift__(self, other):
        if not isinstance(other, _UNIT_TYPES):
            return NotImplemented
        r = _copy.copy(self)
        r.before = [other] + self.before
        return r



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m
import mididings.event as _event
from mididings.extra.per_channel import PerChannel as _PerChannel


class _SustainToNoteoff(object):
    def __init__(self, ctrl):
        self.ctrl = ctrl
        self.pedal = False
        self.notes = set()

    def __call__(self, ev):
        if ev.type == _m.CTRL and ev.ctrl == self.ctrl:
            self.pedal = (ev.value >= 64)
            if self.pedal:
                # pedal pressed
                return None
            else:
                # pedal released, send note offs for all sustained notes
                r = [_event.NoteOffEvent(ev.port, ev.channel, x, 0)
                        for x in self.notes]
                self.notes.clear()
                return r
        elif ev.type == _m.NOTEON and self.pedal:
            # note on while pedal is held
            if ev.note in self.notes:
                self.notes.remove(ev.note)
                return [_event.NoteOffEvent(ev.port, ev.channel, ev.note, 0),
                        ev]
            else:
                return ev
        elif ev.type == _m.NOTEOFF and self.pedal:
            # delay note off until pedal released
            self.notes.add(ev.note)
            return None
        else:
            # everything else: return as is
            return ev


class _SostenutoToNoteoff(object):
    def __init__(self, ctrl):
        self.ctrl = ctrl
        self.pedal = False
        self.held_notes = set()
        self.sustained_notes = set()

    def __call__(self, ev):
        if ev.type == _m.CTRL and ev.ctrl == self.ctrl:
            self.pedal = (ev.value >= 64)
            if self.pedal:
                # pedal pressed, remember currently held notes
                self.sustained_notes |= self.held_notes
                return None
            else:
                # pedal released, send note offs for all sustained notes
                r = [_event.NoteOffEvent(ev.port, ev.channel, x, 0)
                        for x in self.sustained_notes
                            if x not in self.held_notes]
                self.sustained_notes.clear()
                return r
        elif ev.type == _m.NOTEON:
            self.held_notes.add(ev.note)
            return ev
        elif ev.type == _m.NOTEOFF:
            self.held_notes.discard(ev.note)

            # send note off only if the note is not currently being sustained
            if ev.note not in self.sustained_notes:
                return ev
            else:
                return None
        else:
            # everything else: return as is
            return ev


def PedalToNoteoff(ctrl=64, sostenuto=False):
    """
    Convert sustain pedal control changes to note-off events,
    by delaying note-offs until the pedal is released.

    :param ctrl:
        The pedal's controller number.

    :param sostenuto:
        If true act like a sostenuto pedal, instead of a regular sustain
        pedal.
    """
    if sostenuto:
        proc = _m.Process(_PerChannel(lambda: _SostenutoToNoteoff(ctrl)))
    else:
        proc = _m.Process(_PerChannel(lambda: _SustainToNoteoff(ctrl)))

    return (_m.Filter(_m.NOTE) | _m.CtrlFilter(ctrl)) % proc



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

from mididings.units.base import _Filter

import mididings.util as _util
import mididings.overload as _overload
import mididings.arguments as _arguments
import mididings.unitrepr as _unitrepr

import functools as _functools


@_unitrepr.accept(_arguments.flatten(_util.port_number), add_varargs=True)
def PortFilter(ports):
    """
    PortFilter(ports, ...)

    Filter events by port name or number. The *ports* argument can be a single
    port or a list of multiple ports.
    """
    return _Filter(_mididings.PortFilter(map(_util.actual, ports)))


@_unitrepr.accept(_arguments.flatten(_util.channel_number), add_varargs=True)
def ChannelFilter(channels):
    """
    ChannelFilter(channels, ...)

    Filter events by channel number. The *channels* argument can be a single
    channel number or a list of multiple channel numbers.
    System events (which don't carry channel information) are discarded.
    """
    return _Filter(_mididings.ChannelFilter(map(_util.actual, channels)))


@_overload.mark(
    """
    KeyFilter(note_range)
    KeyFilter(lower, upper)
    KeyFilter(lower=...)
    KeyFilter(upper=...)
    KeyFilter(notes=...)

    Filter note events by key (note number or note range). All other events
    are let through.
    The last form expects its argument to be a list of individual note names
    or numbers.
    """
)
@_unitrepr.accept(_util.note_range)
def KeyFilter(note_range):
    return _Filter(_mididings.KeyFilter(note_range[0], note_range[1], []))

@_overload.mark
@_unitrepr.accept(_util.note_limit, _util.note_limit)
def KeyFilter(lower, upper):
    return _Filter(_mididings.KeyFilter(lower, upper, []))

@_overload.mark
@_unitrepr.accept(_util.note_limit)
def KeyFilter(lower):
    return _Filter(_mididings.KeyFilter(lower, 0, []))

@_overload.mark
@_unitrepr.accept(_util.note_limit)
def KeyFilter(upper):
    return _Filter(_mididings.KeyFilter(0, upper, []))

@_overload.mark
@_unitrepr.accept([_util.note_number])
def KeyFilter(notes):
    return _Filter(_mididings.KeyFilter(0, 0, notes))


@_overload.mark(
    """
    VelocityFilter(value)
    VelocityFilter(lower=...)
    VelocityFilter(upper=...)
    VelocityFilter(lower, upper)

    Filter note-on events by velocity. All other events are let through.
    """
)
@_unitrepr.accept(_util.velocity_value)
def VelocityFilter(value):
    return _Filter(_mididings.VelocityFilter(value, value + 1))

@_overload.mark
@_unitrepr.accept(_util.velocity_limit)
def VelocityFilter(lower):
    return _Filter(_mididings.VelocityFilter(lower, 0))

@_overload.mark
@_unitrepr.accept(_util.velocity_limit)
def VelocityFilter(upper):
    return _Filter(_mididings.VelocityFilter(0, upper))

@_overload.mark
@_unitrepr.accept(_util.velocity_limit, _util.velocity_limit)
def VelocityFilter(lower, upper):
    return _Filter(_mididings.VelocityFilter(lower, upper))


@_unitrepr.accept(_arguments.flatten(_util.ctrl_number), add_varargs=True)
def CtrlFilter(ctrls):
    """
    CtrlFilter(ctrls, ...)

    Filter control change events by controller number. The *ctrls* argument
    can be a single controller or a list of multiple controller numbers.
    All other events are discarded.
    """
    return _Filter(_mididings.CtrlFilter(ctrls))


@_overload.mark(
    """
    CtrlValueFilter(value)
    CtrlValueFilter(lower=...)
    CtrlValueFilter(upper=...)
    CtrlValueFilter(lower, upper)

    Filter control change events by controller value. All other events are
    discarded.
    """
)
@_unitrepr.accept(_util.ctrl_value)
def CtrlValueFilter(value):
    return _Filter(_mididings.CtrlValueFilter(value, value + 1))

@_overload.mark
@_unitrepr.accept(_util.ctrl_limit)
def CtrlValueFilter(lower):
    return _Filter(_mididings.CtrlValueFilter(lower, 0))

@_overload.mark
@_unitrepr.accept(_util.ctrl_limit)
def CtrlValueFilter(upper):
    return _Filter(_mididings.CtrlValueFilter(0, upper))

@_overload.mark
@_unitrepr.accept(_util.ctrl_limit, _util.ctrl_limit)
def CtrlValueFilter(lower, upper):
    return _Filter(_mididings.CtrlValueFilter(lower, upper))


@_unitrepr.accept(_arguments.flatten(_util.program_number), add_varargs=True)
def ProgramFilter(programs):
    """
    ProgramFilter(programs, ...)

    Filter program change events by program number. The *programs* argument
    can be a single program number or a list of program numbers.
    All other events are discarded.
    """
    return _Filter(_mididings.ProgramFilter(map(_util.actual, programs)))


@_overload.mark(
    r"""
    SysExFilter(sysex)
    SysExFilter(manufacturer=...)

    Filter system exclusive events by the data they contain, specified as a
    string or as a sequence of integers.
    If *sysex* does not end with ``F7``, partial matches that start with the
    given data bytes are accepted.

    Alternatively, a sysex manufacturer id can be specified, which may be
    either a string or a sequence of integers, with a length of one or three
    bytes.

    All non-sysex events are discarded.
    """
)
@_unitrepr.accept(_functools.partial(_util.sysex_data, allow_partial=True))
def SysExFilter(sysex):
    partial = (sysex[-1] != '\xf7')
    return _Filter(_mididings.SysExFilter(sysex, partial))

@_overload.mark
@_unitrepr.accept(_util.sysex_manufacturer)
def SysExFilter(manufacturer):
    sysex = _util.sysex_to_bytearray([0xf0]) + manufacturer
    return _Filter(_mididings.SysExFilter(sysex, True))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings import overload
from mididings import arguments
from mididings import misc

import sys

import decorator


@decorator.decorator
def store(f, *args, **kwargs):
    """
    Decorator that modifies the function f to store its own arguments in the
    unit it returns.
    """
    unit = f(*args, **kwargs)

    # store the unit's name, the function object, and the value of
    # each argument
    unit._name = f.name if isinstance(f, overload._Overload) else f.__name__
    unit._function = f
    unit._args = args

    return unit


def accept(*constraints, **kwargs):
    """
    @arguments.accept() and @unitrepr.store composed into a single decorator
    for convenience.
    """
    def composed(f):
        return arguments.accept(*constraints, **kwargs) (store(f))
    return composed



def unit_to_string(unit):
    # can't do anything for units that didn't go through @store (or @accept)
    assert hasattr(unit, '_name')

    argnames = misc.getargspec(unit._function)[0]

    args = [
        misc.bytestring(a) if isinstance(a, bytearray) else a
        for a in unit._args
    ]

    # (ab)use misc.formatargspec module to format the arguments used
    formatted = misc.formatargspec(args=argnames, defaults=args)
    return unit._name + formatted


def chain_to_string(chain):
    return ' >> '.join(repr(u) for u in chain)


def fork_to_string(fork):
    if fork.remove_duplicates is not None:
        args = ('units', 'remove_duplicates')
        defaults = (list(fork), fork.remove_duplicates)
        formatted = misc.formatargspec(args, defaults=defaults)
        return 'Fork' + formatted
    else:
        return list.__repr__(fork)


def split_to_string(split):
    return dict.__repr__(split)


def inverted_filter_to_string(invfilt):
    prefix = '-' if invfilt.negate else '~'
    return f"{prefix}{invfilt.filt!r}"



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from mididings import misc

import os
import inspect
import functools


def call(args, kwargs, funcs, name=None):
    """
    Search funcs for a function with parameters such that args and kwargs
    can be applied, and call the first suitable function that is found.
    """
    for f in funcs:
        n = len(args)

        # get argument names and the number of default arguments of f
        argspec = misc.getargspec(f)
        names = argspec[0]
        varargs = argspec[1]
        ndef = len(argspec[3]) if argspec[3] else 0

        # names of the default arguments not overridden by positional arguments
        npos = max(len(names) - ndef, n)
        defargs = names[npos:]

        # check if the number of positional arguments fits, and if the
        # remaining parameters can be filled with keyword and default
        # arguments.
        # alternatively, a suitable function with varargs is also accepted.
        if ((n <= len(names) and set(kwargs) | set(defargs) == set(names[n:]))
                or (n >= len(names) and varargs is not None)):
            # call f with all original arguments
            return f(*args, **kwargs)

    # no overload found, generate a comprehensible error message
    if name is None:
        name = inspect.stack()[1][3]

    # format arg spec for each candidate
    formatargspec = lambda f: misc.formatargspec(*misc.getargspec(f))
    candidates = (f"    {name}{formatargspec(f)}" for f in funcs)

    # formatargspec() doesn't seem to care that the first argument mixes
    # asterisks and argument names
    argx = ['*'] * len(args)
    kwargx = ['*'] * len(kwargs)
    formatvalue = lambda v: f"={v}"
    args_used = misc.formatargspec(argx + list(kwargs.keys()),
                                      defaults=kwargx, formatvalue=formatvalue)

    message = (f"no suitable overload found for {name}{args_used}, candidates are:{os.linesep}{os.linesep.join(candidates)}")
    raise TypeError(message)



class _Overload(object):
    """
    Wrapper class for an arbitrary number of overloads.
    """
    def __init__(self, name, docstring):
        self.name = name
        self.docstring = docstring
        self.funcs = []

    def add(self, f):
        self.funcs.append(f)

    def __call__(self, *args, **kwargs):
        return call(args, kwargs, self.funcs, self.name)


# mapping of all overloaded function names to corresponding _Overload objects
_registry = {}


def _register_overload(f, docstring):
    k = (f.__module__, f.__name__)
    # create a new _Overload object if necessary, add function f to it
    if k not in _registry:
        _registry[k] = _Overload(f.__name__, docstring)
    _registry[k].add(f)
    return k


def mark(f, docstring=None):
    """
    Decorator that marks a function as being overloaded.
    """
    if isinstance(f, str):
        # called with docstring argument. return this function to be used as
        # the actual decorator
        return functools.partial(mark, docstring=f)

    k = _register_overload(f, docstring)

    @functools.wraps(f)
    def call_overload(*args, **kwargs):
        return _registry[k](*args, **kwargs)

    # use the overload's docstring if there is one
    if _registry[k].docstring is not None:
        call_overload.__doc__ = _registry[k].docstring

    # return a function that, instead of calling f, calls the _Overload object
    return call_overload


class partial(object):
    """
    Decorator that adds functools.partial objects as overloads for the given
    function.
    """
    def __init__(self, *partial_args):
        self.partial_args = partial_args

    def __call__(self, f):
        k = _register_overload(f, None)

        # register partial objects as overloads
        for part in self.partial_args:
            _registry[k].add(functools.partial(f, *part))

        @functools.wraps(f)
        def call_overload(*args, **kwargs):
            return _registry[k](*args, **kwargs)

        return call_overload



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m
import mididings.util as _util
import mididings.engine as _engine


class _FloatingKeySplitAnalyzer(object):
    """
    Analyzer class that determines the current split point based on the
    given parameters and the notes being played.
    """
    def __init__(self, threshold_lower, threshold_upper,
                 hold_time, margin_lower, margin_upper):
        self.threshold_lower = _util.note_number(threshold_lower)
        self.threshold_upper = _util.note_number(threshold_upper)
        self.margin_lower = margin_lower
        self.margin_upper = margin_upper
        self.hold_time = hold_time

        self.notes_lower = {}
        self.notes_upper = {}

    def __call__(self, ev):
        now = _engine.time()

        # remove old notes from the lists if hold_time has elapsed
        for k, t in list(self.notes_lower.items()):
            if t and t < now - self.hold_time:
                del self.notes_lower[k]
        for k, t in list(self.notes_upper.items()):
            if t and t < now - self.hold_time:
                del self.notes_upper[k]

        # the lower reference point is the highest note played in the region
        # below the split point, but never below the lower threshold by more
        # than the set margin
        lower = self.threshold_lower - self.margin_lower
        if len(self.notes_lower):
            m = max(self.notes_lower)
            lower = max(m, lower)

        # the upper reference point is the lowest note played in the region
        # above the split point, but never above the upper threshold by more
        # than the set margin
        upper = self.threshold_upper + self.margin_upper
        if len(self.notes_upper):
            m = min(self.notes_upper)
            upper = min(m, upper)

        # calculate new threshold as the center between upper and lower
        # reference point, but confined by the given thresholds
        self.threshold = min(max((lower + upper + 1) // 2,
                                 self.threshold_lower),
                             self.threshold_upper)

        if ev.type == _m.NOTEON:
            # add notes to the appropriate list
            if ev.note < self.threshold:
                self.notes_lower[ev.note] = 0
            else:
                self.notes_upper[ev.note] = 0
        elif ev.type == _m.NOTEOFF:
            # mark notes for removal
            if ev.note in self.notes_lower:
                self.notes_lower[ev.note] = now
            if ev.note in self.notes_upper:
                self.notes_upper[ev.note] = now

        return ev


class _FloatingKeySplitFilter(object):
    """
    Filter class that filters events based on the state of the given analyser.
    """
    def __init__(self, analyzer, index):
        self.analyzer = analyzer
        self.index = index

    def __call__(self, ev):
        # the split point can never move past a note that's still being held,
        # so this is valid for both note-on and note-off events
        if (self.index == 0 and ev.note < self.analyzer.threshold or
            self.index == 1 and ev.note >= self.analyzer.threshold):
            return ev
        else:
            return None


def FloatingKeySplit(threshold_lower, threshold_upper,
                     patch_lower, patch_upper,
                     hold_time=1.0,
                     margin_lower=12, margin_upper=12):
    """
    Create a floating split point that moves dynamically depending on what
    you are playing, allowing a region of the keyboard to be shared between
    two split zones.

    :param threshold_lower:
    :param threshold_upper:
        The lowest and highest notes between which the split point is
        allowed to move.

    :param patch_lower:
    :param patch_upper:
        The patch to which notes below/above the split point will be sent.

    :param hold_time:
        How long released notes will still be taken into account when
        determining the split point (in seconds).

    :param margin_lower:
    :param margin_upper:
        How close you must get to the split point before it starts getting
        pushed into the opposite direction (in semitones).
    """
    # create a single analyzer instance
    analyzer = _FloatingKeySplitAnalyzer(threshold_lower, threshold_upper,
                                         hold_time, margin_lower, margin_upper)

    return _m.Split({
        # separate filter instances are needed for both regions, in order to
        # be able to send note events to different patches
        _m.NOTE:
                _m.Process(analyzer) >> [
                    _m.Process(_FloatingKeySplitFilter(analyzer, 0))
                        >> patch_lower,
                    _m.Process(_FloatingKeySplitFilter(analyzer, 1))
                        >> patch_upper,
                ],
        # non-note-events are sent to both patches
        None:   [ patch_lower, patch_upper ],
    })



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from __future__ import absolute_import as _absolute_import

from mididings import Call as _Call

import dbus as _dbus


class _SendDBUS(object):
    def __init__(self, service, path, interface, method, args):
        self.bus = _dbus.SessionBus()
        self.service = service
        self.path = path
        self.interface = interface
        self.method = method
        self.args = args

    def __call__(self, ev):
        obj = self.bus.get_object(self.service, self.path)
        func = obj.get_dbus_method(self.method, self.interface)
        args = ((x(ev) if hasattr(x, '__call__') else x) for x in self.args)
        func(*args)


def SendDBUS(service, path, interface, method, *args):
    """
    Send a DBUS message.
    Instead of a specific value, each data argument may also be a Python
    function that takes a single :class:`~.MidiEvent` parameter, and
    returns the value to be sent.
    """
    return _Call(_SendDBUS(service, path, interface, method, args))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

import mididings.arguments as _arguments
import mididings.misc as _misc

_VALID_BACKENDS = _mididings.available_backends()

_DEFAULT_CONFIG = {
    'backend':          'alsa' if 'alsa' in _VALID_BACKENDS else 'jack',
    'client_name':      'mididings',
    'in_ports':         1,
    'out_ports':        1,
    'data_offset':      1,
    'octave_offset':    2,
    'initial_scene':    None,
    'start_delay':      None,
    'silent':           False,
}


def _default_portname(n, out):
    prefix = 'out_' if out else 'in_'
    return prefix + str(n + get_config('data_offset'))

def _parse_portnames(ports, out):
    if not _misc.issequence(ports):
        return [_default_portname(n, out) for n in range(ports)]

    portnames = []
    for n, port in enumerate(ports):
        if _misc.issequence(port):
            if port[0] is None:
                portnames.append(_default_portname(n, out))
            else:
                portnames.append(port[0])
        else:
            portnames.append(port)
    return portnames


def _parse_port_connections(ports, out):
    if not _misc.issequence(ports):
        return {}

    connections = {}
    for n, port in enumerate(ports):
        if not _misc.issequence(port) or len(port) <= 1 or port[1] is None:
            continue
        portname = port[0]
        if port[0] is None:
            portname = _default_portname(n, out)
        connections[portname] = port[1:]
    return connections


def reset():
    global _config, _config_overridden, _hooks
    _config = _DEFAULT_CONFIG.copy()
    _config_overridden = []
    _hooks = []
    _config_updated()



@_arguments.accept(kwargs = {
    'backend':          tuple(_VALID_BACKENDS),
    'client_name':      str,
    'in_ports':         _arguments.either(
                            _arguments.each(int,
                                _arguments.condition(lambda x: x >= 0)),
                            _arguments.sequenceof(
                                _arguments.either(str, [str])),
                        ),
    'out_ports':        _arguments.either(
                            _arguments.each(int,
                                _arguments.condition(lambda x: x >= 0)),
                            _arguments.sequenceof(
                                _arguments.either(str, [str])),
                        ),
    'data_offset':      (0, 1),
    'octave_offset':    int,
    'initial_scene':    _arguments.either(
                            int,
                            _arguments.each(tuple, [int]),
                            _arguments.each(tuple, [int, int])
                        ),
    'start_delay':      (int, float, type(None)),
    'silent':           bool,
})
def config(**kwargs):
    """
    config(**kwargs)

    Change :ref:`global settings <main-config>`. This should usually be
    called only once at the beginning of the script, before constructing
    any processing units.

    :param \\*\\*kwargs: an arbitrary number of keyword arguments, matching
        the names and values described in section :ref:`main-config`.
    """
    _config_impl(**kwargs)


def _config_impl(override=False, **kwargs):
    for k, v in kwargs.items():
        # everything seems ok, go ahead and change the config
        if override or k not in _config_overridden:
            _config[k] = v
        if override:
            _config_overridden.append(k)
    _config_updated()


def _config_updated():
    global _in_portnames, _out_portnames
    global _in_port_connections, _out_port_connections
    _in_portnames = _parse_portnames(get_config('in_ports'), False)
    _out_portnames = _parse_portnames(get_config('out_ports'), True)
    _in_port_connections = _parse_port_connections(
                                get_config('in_ports'), False)
    _out_port_connections = _parse_port_connections(
                                get_config('out_ports'), True)


def get_config(var):
    return _config[var]


def hook(*args):
    """
    hook(*args)

    Register "hook" objects, that can be used to extend the functionality
    of mididings.
    Hook classes that ship with mididings are described in section
    :ref:`extra-hooks`.

    :param \\*args: an arbitrary number of hook objects.
    """
    _hooks.extend(args)

def get_hooks():
    return _hooks


reset()
